%global aaModelErrorString;

/*====================================================================================*/
/* NAME:  aaModel                                                                     */
/* BY:    sasdhd & sasdzl                                                             */
/* DATE:  February 29, 2012                                                           */
/* Macro to load the following macros                                                 */
/* aa_model_eval:  macro to evaluate models generates fit statistics and summarize    */
/*                 model scores into centile bins, lift, ...                          */
/* aa_model_register: macro used to register a model or create an SPK file            */
/*                                                                                    */
/*------------------------------------------------------------------------------------*/

%macro aaModel;

%put NOTE: %sysfunc(sasmsg(sashelp.aacomp, model_loadingeval_note , NOQUOTE));
/*====================================================================================*/
/* NAME:  aa_model_eval                                                               */
/* BY:    sasdhd & sasdzl                                                             */
/* DATE:  February 27, 2012                                                           */
/*------------------------------------------------------------------------------------*/
/*------------------------------------------------------------------------------------*/
/*  Macros for summarizing probabilities and creating reports                         */
/*    %aa_model_eval    summarize model scores into centile bins, lift, etc...        */
/*------------------------------------------------------------------------------------*/
/* Macros that are subroutines for internal usage only                                */
/*    aa_microbins_ds1_                                                               */
/*    aa_microbins_hpds2_                                                             */
/*    aa_ishp                                                                         */
/*====================================================================================*/
/* Global macro vars declared in one place                                            */
/*====================================================================================*/
%GLOBAL em_dataset em_keytargetname em_keytargetoutput em_keytargetfrom em_keytargetinto
        em_keytargetlevel em_keytargetevent em_keytargetnonevent em_keytargeteventrate ;
/*====================================================================================*/
/* MACRO aa_microbins_hpds2_;                                                    */
/* Create microbins to summarize a numeric variable without a sort                    */
/* Implementation as HPDS2 code to run on grid appliance                              */
/* Not for end users, should be called from aa_model_register                         */
/*====================================================================================*/

%macro aa_hpds2_genruncode(binvar, sumvar, event, v, format, type, freq);
    %put NOTE: Prefix       =   &v.;
    %put NOTE: format       =   &format.;
    %put NOTE: type         =   &type.;
    %put NOTE: freq         =   &freq.;

    /* accumulate binvar values */
    %if "&freq" ne "" %then %do;
        &V.C_TMP_[INDEX]= &V.C_TMP_[INDEX] + "&freq";
    %end;
    %else %do;
        &V.C_TMP_[INDEX]= &V.C_TMP_[INDEX] +1.0;
    %end;

    if ^missing("&binvar") and ^missing("&sumvar") then do;
       %if "&freq" eq "" %then %do;
            %if "&event" eq "" %then %do;
                &V._DIV_TMP_[INDEX]= &V._DIV_TMP_[INDEX]+1;
            %end;
            %else %do;
                &V._NOBS_TMP_[INDEX] = &V._NOBS_TMP_[INDEX] + 1;
            %end;
       %end;
       %else %do;
           if "&freq">0 then do;
              %if "&event" eq "" %then %do;
                  &V._DIV_TMP_[INDEX] = &V._DIV_TMP_[INDEX]+ &freq;
              %end;
              %else %do;
                  &V._NOBS_TMP_[INDEX] = &V._NOBS_TMP_[INDEX] + &freq;
              %end;
           end;
       %end;
    end;

    if ^missing(&binvar) then do;
        &V.P_TMP_[INDEX] = &V.P_TMP_[INDEX] + &binvar;
        &V.M_TMP_[INDEX] = MIN(&V.M_TMP_[INDEX], &binvar);
        &V.X_TMP_[INDEX] = MAX(&V.X_TMP_[INDEX], &binvar);
    end;

    /* accumulate sumvar values */
    if ^missing("&sumvar") then do;
       %if "&freq" ne "" %then %do;
           &V.SC_TMP_[INDEX] = &V.SC_TMP_[INDEX]+ "&freq";
       %end;
       %else %do;
           &V.SC_TMP_[INDEX] = &V.SC_TMP_[INDEX]+ 1.0;
       %end;

       %if "&event" ne "" %then %do;
           %if "&format" eq "" %then %do;
               %if "&type" eq "1" %then %do;
                   if "&sumvar" = &EVENT then do;
                     %end;
                %else %do;
                    if DMNORM("&sumvar") = DMNORM(%STR(%')%nrbquote(&EVENT)%STR(%')) then do;
                %end;
           %end;
           %else %do;
               if DMNORM(PUT("&sumvar", &format)) = DMNORM(%STR(%')%nrbquote(&EVENT)%STR(%')) then do;
           %end;

           %if "&freq" ne "" %then %do;
               if "&freq">0 then do;
                  if ("&binvar"<&cutoff) then
                      &V._MISC_TMP_[INDEX] = &V._MISC_TMP_[INDEX] + "&freq";
                      &V.E_TMP_[INDEX] = &V.E_TMP_[INDEX] + "&freq";
               end;
           %end;
           %else %do;
               if ("&binvar"<&cutoff) then
                  &V._MISC_TMP_[INDEX] = &V._MISC_TMP_[INDEX] + 1;
                  &V.E_TMP_[INDEX] = &V.E_TMP_[INDEX] + 1.0;
           %end;
               end ;
               else do;
                  %if "&freq" ne "" %then %do;
                      if ("&binvar">=&cutoff) and "&freq">0 then
                          &V._MISC_TMP_[INDEX] = &V._MISC_TMP_[INDEX] + "&freq";
                  %end;
                  %else %do;
                      if ("&binvar">=&cutoff) then
                         &V._MISC_TMP_[INDEX] = &V._MISC_TMP_[INDEX] + 1;
                  %end;
               end;
        %end;
        %else %do;
            &V.SP_TMP_[INDEX]= &V.SP_TMP_[INDEX] + "&sumvar";
            &V.SM_TMP_[INDEX]= min(&V.SM_TMP_[INDEX], "&sumvar") ;
            &V.SX_TMP_[INDEX]= max(&V.SX_TMP_[INDEX], "&sumvar") ;
            _R_= ("&binvar" - "&sumvar") ;
            if _R_ ne . then do;
               &V.RM_TMP_[INDEX]= min(&V.RM_TMP_[INDEX], _R_) ;
               &V.RX_TMP_[INDEX]= max(&V.RX_TMP_[INDEX], _R_) ;
               %if "&freq" ne "" %then %do;
                   if "&freq">0 then do;
                      &V.RP_TMP_[INDEX]= &V.RP_TMP_[INDEX] + _R_ * "&freq";
                      &V.RR_TMP_[INDEX]= &V.RR_TMP_[INDEX] + _R_ * _R_ * "&freq";
                   end;
               %end;
               %else %do;
                   &V.RP_TMP_[INDEX]= &V.RP_TMP_[INDEX] + _R_ ;
                   &V.RR_TMP_[INDEX]= &V.RR_TMP_[INDEX] + _R_ * _R_ ;
               %end;
            end;
        %end;
      end;

%mend aa_hpds2_genruncode;

%macro aa_microbins_hpds2_(data,freq,binvar,sumvar,event,deqevent,format, type, maxbins,fuzz,min,max,partitionvar);
    %put NOTE: data         =   &data. ;
    %put NOTE: binvar       =   &binvar.    -- will be binned. ;
    %put NOTE: sumvar       =   &sumvar.    -- will be summarized. ;
    %if "&event" ne " " %then %do ;
        %put NOTE: event    =   %nrbquote(&event)  -- will be counted.  ;
    %end ;

    %put NOTE: type         =   &type.;
    %put NOTE: format       =   &format.;
    %put NOTE: freq         =   &freq.  ;

    %put NOTE: PartitionVar =   &partitionVar.;


    %let binvar = %upcase(&binvar);
    %let sumvar = %upcase(&sumvar);
    %let freq   = %upcase(&freq);

    %local mname ;
    %let mname= (aa_microbins_hpds2_) ;

    %if "&event" ne " " %then %do ;
        data _null_;
            putlog "NOTE: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_eventtocount_note, NOQUOTE,&event))";
        run;
    %end ;

    /* Check if gpcharexpand option needs to be used */
    %let gpdbcs = ;
    %let bpc =;
    %if ^%symexist(hpdm_performance) %then %do;
        %let ENGINE = SAS;
        %if "&data" ne "" %then %do;
            %let dsid = %sysfunc(open(&DATA));
            %if &dsid> 0 %then %do;
                %let ENGINE = %upcase(%sysfunc(attrc(&dsid, ENGINE)));
            %end;
            %if &dsid > 0 %then %let dsid = %sysfunc(close(&dsid));
        %end;
        %if "&ENGINE" = "GREENPLM" %then %do;
            %if ^%symexist(HPDM_GPCHAREXPAND) %then %do;
                %if (%upcase("&sysencoding") ne "LATIN1" and %upcase("&sysencoding") ne "WLATIN1") %then
                    %let gpdbcs = gpcharexpand;
            %end;
            %else %do;
                %if (%upcase("&HPDM_GPCHAREXPAND") = "Y") or (%upcase("&HPDM_GPCHAREXPAND") = "YES") %then %do;
                    %let gpdbcs= gpcharexpand;
                %end;
                %if (%upcase("&HPDM_GPCHAREXPAND") = "N") or (%upcase("&HPDM_GPCHAREXPAND") = "NO") %then %do;
                    %let gpdbcs=;
                %end;
            %end;
        %end;
        %else
            %if %upcase("&ENGINE") = "TERADATA" %then %do;
                %if %symexist(HPDM_BYTESPERCHAR) %then %do;
                    %if %sysfunc(indexc(&HPDM_BYTESPERCHAR, 1)) and ^%length(%sysfunc(compress(&HPDM_BYTESPERCHAR, 1))) %then %do;
                        %let bpc = BYTESPERCHAR=1;
                    %end;
                %end;
             %end;

     %end;

    *----------- get range of variable ----------*;
    *----------- create micro bins ----------*;
    %if %sysfunc(exist(microbins)) %then %do;
        proc delete data=microbins;
        run;
    %end;
    %if %sysfunc(exist(_newassess_stats)) %then %do;
        proc delete data=_newassess_stats;
        run;
    %end;

    %let sumvar = %upcase(&sumvar);
    %let binvar = %upcase(&binvar);

    %let Part=0;
    %let num=1;
    %if "&PartitionVar" ne "" %then %do;
        %let Part=1;
        %let num=2;
    %end;


    %if "&min" eq "" or "&max" eq "" %then %do;
        ods listing close;
        %if ^&part %then %do;
            proc hpdmdb data=&data varout=_minmaxbins;
               var &binvar

               %if "&event" eq "" %then %do;
                   &sumvar.
               %end;

               ;

               %if "&freq" ne "" %then %do;
                   freq &freq;
               %end;

               %if %symexist(hpdm_performance) %then %do;
                   &hpdm_performance;
               %end;
               %else %do;
                   performance details &gpdbcs &bpc;
               %end;
            run;
            data _null_;
               set _minmaxbins;
               if upcase(NAME) = "&binvar" then do;
                  call symput('MIN' , put(MIN,best.)) ;
                  call symput('MAX' , put(MAX,best.)) ;
               end;
               %if "&event" eq "" %then %do;
                   if upcase(NAME) = "&sumvar" then do;
                      call symput('MEAN', put(MEAN,best.)) ;
                      call symput('STD' , put(STD,best.)) ;
                   end;
              %end;
           run;
        %end;
        %else %do;
            proc hpsummary data=&data;
               class &partitionvar;

               var &binvar

               %if "&event" eq "" %then %do;
                   &sumvar.
               %end;

               ;

               %if "&freq" ne "" %then %do;
                   freq &freq;
               %end;
               output out=_minmaxbins;
               %if %symexist(hpdm_performance) %then %do;
                   &hpdm_performance;
               %end;
               %else %do;
                   performance details &gpdbcs &bpc;
               %end;
            run;
            data _null_;
               set _minmaxbins;
               if _Partind_=1 then do;
                  if upcase(_STAT_) = 'MIN' then call symput('MIN' , put(&binvar,best.)) ;
                  if upcase(_STAT_) = 'MAX' then call symput('MAX' , put(&binvar,best.)) ;
                  %if "&event" eq "" %then %do;
                      if upcase(_STAT_) = 'MEAN' then call symput('MEAN', put(&sumvar,best.)) ;
                      if upcase(_STAT_) = 'STD' then call symput('STD'  , put(&sumvar,best.)) ;
                  %end;
               end;
               %if "&event" eq "" %then %do;
                   else do;
                      if upcase(_STAT_) = 'MEAN' then call symput('VMEAN' , put(&sumvar,best.)) ;
                      if upcase(_STAT_) = 'STD'  then call symput('VSTD'  , put(&sumvar,best.)) ;
                   end;
               %end;
            run;

        %end;
        ods listing;

        proc delete data=_minmaxbins;
        run;
    %end;
    %if %sysevalf(%sysfunc(sum(&MIN.,&MAX.))= 0) %then %do ;
        %let FUZZ=1 ;
        data _null_;
           putlog "WARNING: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_minequalmax_warning, NOQUOTE,&fuzz))";
        run;
    %end ;
    %if "&event" eq "" %then %do ;
        data _null_;
            putlog "NOTE: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_summarize_note, NOQUOTE,&binvar, &min, &max))";
            putlog "NOTE: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_summarizenoevent_note, NOQUOTE))";
        run;
    %end ;
    %else %do ;
        data _null_;
           putlog "NOTE: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_summarize_note, NOQUOTE,&binvar, &min, &max))";
           putlog "NOTE: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_summarizeevent_note, NOQUOTE, &event, &sumvar))";
        run;
    %end ;

    %let cutoff = 0.5;
    %if %symexist(aa_prob_cutoff) %then %do;
        %let cutoff = &aa_prob_cutoff;
        %if %sysevalf(&cutoff<=0) or %sysevalf(1-&cutoff<=0) %then
            %let cutoff=0.5;
    %end;


    proc hpds2 in=&data
       out=microbins(keep=

       %do i= 1 %to &num;
           %let V=;
           %if "&i" = "2" %then
                %let V = V;

           &v.COUNT &v.SCOUNT

           %if "&event" ne "" %then %do;
                &v.EVENTS _&v.MISC_ _&v.NOBS_
           %end;
           %else %do;
               &v.SMEAN &v.SMIN &v.SMAX &v.RMEAN &v.RMIN &v.RMAX &v.RRMEAN  _&v.DIV_ _&v.SSE_
           %end;

           &v.MEAN &v.MIN &v.MAX
       %end;

       );

       %if %symexist(hpdm_performance) %then %do;
           &hpdm_performance;
       %end;
       %else %do;
           performance details &gpdbcs &bpc;
       %end;


    data DS2GTF.out;
       dcl double I;

   %do i= 1 %to &num;
       %let V=;
        %if "&i" = "2" %then
            %let V = V;


       dcl double &V.INDEX;
       dcl double &v.COUNT;
       dcl double &v.SCOUNT;
       dcl double &v.MEAN;
       dcl double &v.MIN;
       dcl double &v.MAX;

       %if "&event" eq "" %then %do;
           dcl double _&V.R_;
           dcl double &V.SMEAN;
           dcl double &V.SMIN;
           dcl double &V.SMAX;
           dcl double &V.SMEAN;
           dcl double &V.RMEAN;
           dcl double &V.RMIN;
           dcl double &V.RMAX;
           dcl double &V.RRMEAN;
           dcl double &V._DIV_TMP_[&maxbins.];
       %end;
       %else %do;
           dcl double &v._MISC_TMP_[&maxbins.];
           dcl double &v._NOBS_TMP_[&maxbins.];
       %end;

       %if "&event" ne "" %then %do ;
           dcl double &V.EVENTS;
           dcl double &V.E_TMP_[&maxbins];
       %end;
       %else %do;
           dcl double &V.SP_TMP_[&maxbins.];
           dcl double &V.SX_TMP_[&maxbins.];
           dcl double &V.SM_TMP_[&maxbins.];
           dcl double &V.RP_TMP_[&maxbins.];
           dcl double &V.RM_TMP_[&maxbins.];
           dcl double &V.RX_TMP_[&maxbins.];
           dcl double &V.RR_TMP_[&maxbins.];
       %end;

       dcl double &V.C_TMP_[&maxbins];  /* total counts      */
       dcl double &V.P_TMP_[&maxbins];  /* numeric values    */
       dcl double &V.M_TMP_[&maxbins];  /* min values        */
       dcl double &V.X_TMP_[&maxbins];  /* max values        */
       dcl double &V.SC_TMP_[&maxbins]; /* sumvar counts     */
 %end;

       retain MIN 0 MAX 0;
       method init();
            do I = 1.0 to &maxbins..0;

       %do i= 1 %to &num;
           %let V=;
           %if "&i" = "2" %then
                %let V = V;

               %if "&event" ne "" %then %do;
                   /*--- collect counts of events ---*/
                   &V.E_TMP_[I] = 0.0;
                   &V._NOBS_TMP_[I] = 0.0;
                   &V._MISC_TMP_[I] = 0.0;
               %end;
               %else %do;
                   /*--- collect stats for interval summary var ---*/
                   &V.SP_TMP_[I]=0;
                   &V.SM_TMP_[I]=1E16;
                   &V.SX_TMP_[I]=-1E16;
                   &V.RP_TMP_[I]=0;
                   &V.RM_TMP_[I]=1E16;
                   &V.RX_TMP_[I]=-1E16;
                   &V.RR_TMP_[I]=0;
                   &V._DIV_TMP_[I] = 0.0;
               %end;

               &V.C_TMP_[I]  = 0.0;
               &V.SC_TMP_[I] = 0.0;
               &V.P_TMP_[I]  = 0.0;
               &V.M_TMP_[I]  = 1E16;
               &V.X_TMP_[I]  = -1E16;

       %end;
               MIN = &min;
               MAX = &max;
            end;
      end;
      method run();
           set DS2GTF.in;
           /* reserve bin 1 for missing values counting var */
           if ^missing(&binvar) then do;
              %if &FUZZ %then %do;
                  INDEX = 2.0 + (&MAXBINS - 4.0) * ((&binvar - MIN) / (MAX - MIN + 1E-12) + RANNOR(1.0) / 64.0);
              %end;
              %else %do;
                  INDEX = 2.0 + (&MAXBINS - 4.0) * ((&binvar - MIN) / (MAX - MIN + 1E-12));
              %end;
              INDEX = floor(MIN(&MAXBINS, MAX(2.0, INDEX)));
          end;
          else do;
             INDEX=1.0;
          end;

          /* accumulate binvar values */
          %let v = ;
          %if &Part %then %do;
              if "&partitionVar" = 1 then do;
                 %aa_hpds2_genruncode(&binvar, &sumvar, %nrbquote(&deqevent), &v, &format, &type, &freq); /* Generate Train Code */
              end;
              else
              if "&PartitionVar" = 0 then do;
                 %let v = V;
                 %aa_hpds2_genruncode(&binvar, &sumvar, %nrbquote(&deqevent), &v, &format, &type, &freq); /* Generate Validate Code */
              end;
          %end;
          %else %do;
              %aa_hpds2_genruncode(&binvar, &sumvar, %nrbquote(&deqevent), &v,&format, &type, &freq);
          %end;

      end;

      /* output result */
      method term();
       %do i= 1 %to &num;
           %let V=;
           %if "&i" = "2" %then
                %let V = V;

          &v.COUNT  = 0.0;
          &v.SCOUNT = 0.0;
          &v.MIN    = 0.0;
          &v.MAX    = 0.0;
          %if "&event" ne "" %then %do;
              &v.EVENTS = 0.0;
              _&v.MISC_  =0.0;
              _&v.NOBS_  =0.0;
          %end;
          %else %do;
              &v.SMEAN  = 0.0;
              &v.SMIN   = 0.0;
              &v.SMAX   = 0.0;
              &v.RMEAN  = 0.0;
              &v.RMIN   = 0.0;
              &v.RMAX   = 0.0;
              &v.RRMEAN = 0.0;
              _&v.DIV_  = 0.0;
              _&v.SSE_  = 0.0;
          %end;
       %end;

          do I = 1.0 to &MAXBINS;
             %do i= 1 %to &num;
                 %let V=;
                 %if "&i" = "2" %then
                     %let V = V;

                &v.COUNT  = &v.C_TMP_[I];
                &v.SCOUNT = &v.SC_TMP_[I];
             %end;

             %let vstring = ;
             %if &part %then %let vstring = or VCOUNT>0 or VSCOUNT>0;

             if COUNT>0 or SCOUNT>0 &vstring then do ; /* write a record to microbins */

             %do i= 1 %to &num;
                 %let V=;
                 %if "&i" = "2" %then
                     %let V = V;

                if I>1.0 and &v.COUNT>=1.0 then do;
                   &v.MEAN = &v.P_TMP_[I] / &v.COUNT;
                   &v.MIN  = &v.M_TMP_[I];
                   &v.MAX  = &v.X_TMP_[I];
                end;

                %if "&event" ne "" %then %do;
                    &v.EVENTS = &v.E_TMP_[I];
                    _&v.NOBS_ = &v._NOBS_TMP_[I];
                    _&v.MISC_ = &v._MISC_TMP_[I];
                %end;
                %else %do;
                    if ^missing(&v.RR_TMP_[I]) then _&v.SSE_  = &v.RR_TMP_[I];
                    _&v.DIV_ = &v._DIV_TMP_[I];

                    if &v.SCOUNT>=1.0 then do;
                       &v.SMEAN=  &v.SP_TMP_[I]/&v.SCOUNT ;
                       &v.SMIN=   &v.SM_TMP_[I] ;
                       &v.SMAX=   &v.SX_TMP_[I] ;
                       &v.RMEAN=  &v.RP_TMP_[I]/&v.SCOUNT ;
                       &v.RMIN=   &v.RM_TMP_[I] ;
                       &v.RMAX=   &v.RX_TMP_[I] ;
                       &v.RRMEAN= &v.RR_TMP_[I]/&v.SCOUNT ;
                    end;
                %end;
             %end;
                output;
             end;
          end;
      end;
   enddata;
   run;
   quit;

   %if %sysfunc(exist(microbins)) %then %do;
       proc dmdb data=microbins varout=_newassess_stats;
          var

          %do i= 1 %to &num;
                 %let V=;
                 %if "&i" = "2" %then
                     %let V = V;


              %if "&event" ne "" %then %do;
                  _&v.MISC_ _&v.NOBS_
              %end;
              %else %do;
                  _&v.DIV_ _&v.SSE_
              %end;
          %end;
          ;

          where MEAN ne .;
       run;
       data _newassess_stats;
          set _newassess_stats end=eof;
          %if "&event" ne "" %then %do;
              keep _MISC_;
              retain _MISC_ _NOBS_ 0;

              %if &part %then %do;
                  keep _VMISC_;
                  retain _VMISC_ _VNOBS_ 0;
              %end;

              if      NAME = '_MISC_' then _MISC_=MEAN*N;
              else if NAME ='_NOBS_'  then _NOBS_=MEAN*N;
              %if &part %then %do;
                  else if NAME = '_VMISC_' then _VMISC_=MEAN*N;
                  else if NAME='_VNOBS_' then _VNOBS_=MEAN*N;
              %end;

              if eof then do;
                 if _NOBS_ then _MISC_=_MISC_/_NOBS_;
                else _MISC_ = .;
                %if &part %then %do;
                    if _VNOBS_ then  _VMISC_ = _VMISC_/_VNOBS_;
                    else _VMISC_ = .;
                 %end;

                 output;
              end;
          %end;
          %else %do;
              keep _DIV_ _SSE_ _ASE_ _RASE_ _MEANP_ _STDP_ _CVP_;
              retain _DIV_ 0 _SSE_ 0;
              %if &part %then %do;
                  keep _VDIV_ _VSSE_ _VASE_ _VRASE_ _VMEANP_ _VSTDP_ _VCVP_;
                  retain _VDIV_ 0 _VSSE_ 0;
              %end;

             if NAME = '_DIV_' then _DIV_=MEAN*N;
              else
              if NAME='_SSE_' then _SSE_=MEAN*N;
              %if &part %then %do;
                 else
                 if NAME = '_VDIV_' then  _VDIV_ =MEAN*N;
                 else if NAME='_VSSE_' then _VSSE_=MEAN*N;
             %end;

             if eof then do;
                 if _DIV_ ne 0 then do;
                    _ASE_  = _SSE_/_DIV_;
                    _RASE_ = sqrt(_ASE_);
                 end;
                 else do;
                    _ASE_ = .;
                    _RASE_ = .;
                 end;

                %if &part %then %do;
                    if _VDIV_ ne 0 then do;
                       _VASE_  = _VSSE_/_VDIV_;
                       _VRASE_ = sqrt(_VASE_);
                    end;
                    else do;
                       _VASE_ = .;
                       _VRASE_ = .;
                    end;
                %end;

                 _MEANP_=&MEAN;
                 _STDP_=&STD;
                 if _MEANP_ ne 0 then _CVP_= _STDP_/_MEANP_;
                 %if &part %then %do;
                     _VMEANP_=&vMEAN;
                     _VSTDP_=&vSTD;
                     if _VMEANP_ ne 0 then _VCVP_= _VSTDP_/_vMEANP_;

                 %end;
                 output;
             end;
        %end;
     run;
 %end;

%mend aa_microbins_hpds2_;
/*====================================================================================*/
/* MACRO aa_microbins_ds1_;                                                           */
/* Create microbins to summarize a numeric variable without a sort                    */
/* Implementation as SAS code to run on SAS system                                    */
/* Not for end users, should be called from aa_model_register                         */
/* binvar will be used to drive the binning                                           */
/* sumvar will be used to drive the summary                                           */
/*====================================================================================*/
%macro aa_microbins_ds1_(data,freq,binvar,sumvar,event,format,type, maxbins,fuzz,min,max, partitionvar);
    %local mname ;
    %let mname= (aa_microbins_sas_) ;
    %put NOTE: &mname ;
    %put NOTE: data     =   &data. ;
    %put NOTE: binvar   =   &binvar.    -- will be binned. ;
    %put NOTE: sumvar   =   &sumvar.    -- will be summarized. ;
    %if "&event." ne "" %then %do ;
         %put NOTE: event    =   &event.     -- will be counted.  ;
    %end ;
    %put NOTE: format   =   &format.;
    %put NOTE: type     =   &type.;
    %put NOTE: freq     =   &freq.  ;
    %put NOTE: PartitionVar =   &partitionVar;

    %let Part=1;
    %if "&PartitionVar" ne "" %then %do;
        %let Part=2;

        %let aaPartitionVar = %upcase(&PartitionVar);
    %end;

    /*----------- get range of variable ----------*/;
    %if "&min." eq "" or "&max." eq "" %then %do  ;
        proc means data=&data min max mean noprint ;output out=_aamodelMean ;var &binvar.

             %if "&event" eq "" %then %do;
                 &sumvar.
             %end;
                   ;

              %if "&PartitionVar" ne "" %then %do;
                 class &PartitionVar;
             %end;

             %if "&freq" ne "" %then %do;
                 freq &freq;
             %end;

        run ;

        data _null_;set _aamodelMean;
            %if "&PartitionVar" eq "" %then %do;
                if _stat_ eq "MIN" then call symput('MIN' , put(&binvar.,best.)) ;
                if _stat_ eq "MAX" then call symput('MAX' , put(&binvar.,best.)) ;
                %if "&event" eq "" %then %do;
                    if _stat_ eq "MEAN" then call symput('MEAN' , put(&sumvar.,best.));
                    if _stat_ eq "STD" then call symput('STD' ,   put(&sumvar.,best.));
                %end;
            %end;
            %else %do;
              if &PartitionVar=1 and _stat_ eq "MIN" then call symput('MIN' , put(&binvar.,best.)) ;
              if &PartitionVar=1 and _stat_ eq "MAX" then call symput('MAX' , put(&binvar.,best.)) ;
              %if "&event" eq "" %then %do;
                  if &PartitionVar=1 then do;
                     if _stat_ eq "MEAN" then call symput('MEAN' , put(&sumvar.,best.));
                     if _stat_ eq "STD"  then call symput('STD' ,   put(&sumvar.,best.));
                  end;
                  else do;
                     if _stat_ eq "MEAN" then call symput('VMEAN' , put(&sumvar.,best.));
                     if _stat_ eq "STD"  then call symput('VSTD' ,   put(&sumvar.,best.));
                  end;
              %end;
          %end;
        run ;
        proc delete data=_aamodelMean; run;
    %end ;
    %if %sysevalf(%sysfunc(sum(&MIN.,&MAX.))= 0) %then %do ;
        %let FUZZ=1 ;
        data _null_;
           putlog "WARNING: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_minequalmax_warning, NOQUOTE,&fuzz))";
        run;
    %end ;
    %if "&event" eq "" %then %do ;
        data _null_;
            putlog "NOTE: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_summarize_note, NOQUOTE,&binvar, &min, &max))";
            putlog "NOTE: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_summarizenoevent_note, NOQUOTE))";
        run;
    %end ;
    %else %do ;
        data _null_;
           putlog "NOTE: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_summarize_note, NOQUOTE,&binvar, &min, &max))";
           putlog "NOTE: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_summarizeevent_note, NOQUOTE, &event, &sumvar))";
        run;
    %end ;

    %let cutoff = 0.5;
    %if %symexist(aa_prob_cutoff) %then %do;
        %let cutoff = &aa_prob_cutoff;
        %if %sysevalf(&cutoff<=0) or %sysevalf(1-&cutoff<=0) %then
            %let cutoff=0.5;
    %end;


    /*----------- create micro bins ----------*/;
    data microbins _newassess_stats;
       set &data. end=eof;
       length index 8;
   %do i=1 %to &part;
       %let V = ;
       %if "&i" = "2" %then %do;
           %let V = V;
       %end;

           length &V.count &V.scount &V.mean &V.min &V.max 8;
       %if "&event" eq "" %then %do ;
           length _&V.sse_ _&V.div_ 8 ;
       %end;
       %else %do;
           length _&V.misc_ _&V.nobs_ 8;
       %end;
       keep &V.count &V.scount &V.mean &V.min &V.max

       %if "&event" eq "" %then %do ;
           _&V.sse_ _&V.div_
       %end;
       %else %do;
           _&V.misc_ _&V.nobs_
       %end;

       ;

        /*--- binning measures for bin var in all cases --------*/
        /*--- class case: this will summarize probabilities  ---*/
        /*--- interval case: this will summarize the sumvar  ---*/
        array &V.c_tmp_[&maxbins.] 8 _TEMPORARY_ ;  /* total counts      */
        array &V.p_tmp_[&maxbins.] 8 _TEMPORARY_ ;  /* numeric values    */
        array &V.m_tmp_[&maxbins.] 8 _TEMPORARY_ ;  /* min values        */
        array &V.x_tmp_[&maxbins.] 8 _TEMPORARY_ ;  /* max values        */
        array &V.sc_tmp_[&maxbins.] 8 _TEMPORARY_;  /* sumvar counts     */

        retain &V.min 0 &V.max 0 ;
        %if "&event" eq "" %then %do ;
            retain _&V.div_ 0;
        %end;
        %else %do;
                retain _&V.misc_ 0 _&V.nobs_ 0;
        %end;

        if _n_ eq 1 then do i=1 to &maxbins. ;
            &V.c_tmp_[i]=0; &V.sc_tmp_[i]=0;&V.p_tmp_[i]=0;&V.m_tmp_[i]=1e16;&V.x_tmp_[i]=-1e16;
            &V.min= &min. ; &V.max= &max. ;
        end ;
        /*--- now initialize summary data ---*/
        %if "&event" ne "" %then %do ;
            /*--- collect counts of events ---*/
            length &V.events  8 ;
            keep &V.events ;
            array &V.E_tmp_[&maxbins.] 8 _TEMPORARY_ ; /* event counts      */
            if _n_ eq 1 then do i=1 to &maxbins. ; &V.e_tmp_[i]=0;end ;
        %end;
        %else %do;
            /*--- collect stats for interval summary var ---*/
            length  &V.smean &V.smin &V.smax &V.rmean &V.rmin &V.rmax &V.rrmean 8  ;
            keep  &V.smean &V.smin &V.smax &V.rmean &V.rmin &V.rmax &V.rrmean ;
            array &V.sp_tmp_[&maxbins.] 8 _TEMPORARY_ ; /* numeric values    */
            array &V.sm_tmp_[&maxbins.] 8 _TEMPORARY_ ; /* min values        */
            array &V.sx_tmp_[&maxbins.] 8 _TEMPORARY_ ; /* max values        */
            array &V.rp_tmp_[&maxbins.] 8 _TEMPORARY_ ; /* differences       */
            array &V.rm_tmp_[&maxbins.] 8 _TEMPORARY_ ; /* min values        */
            array &V.rx_tmp_[&maxbins.] 8 _TEMPORARY_ ; /* max values        */
            array &V.rr_tmp_[&maxbins.] 8 _TEMPORARY_ ; /* squares           */
            if _n_ eq 1 then do i=1 to &maxbins. ;
                &V.sp_tmp_[i]=0;&V.sm_tmp_[i]=1e16;&V.sx_tmp_[i]=-1e16;
                &V.rp_tmp_[i]=0;&V.rm_tmp_[i]=1e16;&V.rx_tmp_[i]=-1e16;&V.rr_tmp_[i]=0;
            end ;
        %end ;
    %end;

        /* reserve bin 1 for missing values counting var */
        if missing(&binvar.) then do ;
            index= 1 ;
        end ;
        else do ;
            %if &fuzz ne 0 %then %do ;
                index= 2+(&maxbins.-4)*((&binvar.-min) / (max-min+1e-12) + rannor(1)/64 ) ;
            %end;%else %do ;
                index= 2+(&maxbins.-4)*(&binvar.-min) / (max-min+1e-12)  ;
            %end ;
            index= min(&maxbins.,max(2,index));
        end;

     %do i=1 %to &part;
       %let V = ;
       %if &i = 2 %then %let V = V;
         %if "&PartitionVar" ne "" %then %do;
             %if &i=1 %then %do;
                 if &PartitionVar = 1 then do;
             %end;
             %else %do;
                 if &PartitionVar = 0 then do;
             %end;
        %end;

        if &binvar. ne . and &sumvar. ne . then do;
         %if "&freq" eq "" %then %do;
              _&V.div_+1;
         %end;
         %else %do;
             if &freq>0 then do;
                _&V.div_ + &freq;
             end;
         %end;
      end;


        * accumulate binvar values ;
        %if "&freq" ne "" %then %do;
              &V.c_tmp_[index] = &V.c_tmp_[index]+ &freq;
            %end;
            %else %do;
              &V.c_tmp_[index] = &V.c_tmp_[index]+ 1;
        %end;
        if not missing(&binvar.) then  do ;
            &V.p_tmp_[index]= &V.p_tmp_[index] + &binvar. ;
            &V.m_tmp_[index]= min(&V.m_tmp_[index], &binvar.) ;
            &V.x_tmp_[index]= max(&V.x_tmp_[index], &binvar.) ;
        end ;
        * accumulate sumvar values ;
        if not missing(&sumvar.) then do ;
           %if "&freq" ne "" %then %do;
               &V.sc_tmp_[index] = &V.sc_tmp_[index]+ &freq;
           %end;
           %else %do;
               &V.sc_tmp_[index] = &V.sc_tmp_[index]+ 1;
           %end;
           %if "&event" ne "" %then %do ;
                %if "&freq" ne "" %then %do;
                    if &freq>0 then _&V.nobs_ = _&V.nobs_ + &freq;
                %end;
                %else %do;
                    _&V.nobs_ = _&V.nobs_ + 1;
                %end;

                %if "&format" ne "" %then %do;
                    if dmnorm(put(&sumvar.,&format.)) eq dmnorm("&event.") then do ;
                %end;
                %else %do;
                    %if "&type" eq "1" %then %do;
                        if &sumvar. = &event. then do ;
                    %end;
                    %else %do;
                        if dmnorm(&sumvar.) eq dmnorm("&event.") then do ;
                     %end;
                 %end;
                 %if "&freq" ne "" %then %do;
                     if &freq>0 then do;
                        if (&binvar<&cutoff) then
                            _&V.misc_ = _&V.misc_ + &freq;
                         &V.e_tmp_[index] = &V.e_tmp_[index]+ &freq;
                      end;
                 %end;
                 %else %do;
                     if (&binvar<&cutoff) then
                        _&V.misc_ = _&V.misc_ + 1;
                     &V.e_tmp_[index] = &V.e_tmp_[index]+ 1;
                 %end;
                 end ;
                 else do;
                    %if "&freq" ne "" %then %do;
                        if (&binvar>=&cutoff) and &freq>0 then
                           _&V.misc_ = _&V.misc_ + &freq;
                    %end;
                    %else %do;
                        if (&binvar>=&cutoff) then
                           _&V.misc_ = _&V.misc_ + 1;
                    %end;
                 end;
             %end ;
             %else %do ;
                &V.sp_tmp_[index]= &V.sp_tmp_[index] + &sumvar. ;
                &V.sm_tmp_[index]= min(&v.sm_tmp_[index], &sumvar.) ;
                &V.sx_tmp_[index]= max(&v.sx_tmp_[index], &sumvar.) ;
                _r_= (&binvar. - &sumvar.) ;
                if _r_ ne . then do;
                   &V.rm_tmp_[index]= min(&V.rm_tmp_[index], _r_) ;
                   &V.rx_tmp_[index]= max(&V.rx_tmp_[index], _r_) ;
                   %if "&freq" ne "" %then %do;
                       if &freq>0 then do;
                          &V.rp_tmp_[index]= &V.rp_tmp_[index] + _r_ * &freq;
                          &V.rr_tmp_[index]= &V.rr_tmp_[index] + _r_ * _r_ * &freq;
                       end;
                   %end;
                   %else %do;
                       &V.rp_tmp_[index]= &V.rp_tmp_[index] + _r_;
                       &V.rr_tmp_[index]= &V.rr_tmp_[index] + _r_ * _r_;
                    %end;
                end;
            %end;

        %if "&partitionvar" ne "" %then %do;
             end ;
        %end;

        end;

       %end;

        /* output result */
        if eof then do ;
           %do i=1 %to &part;
             %let V = ;
             %if &i = 2 %then %let V = V;

              _&V.sse_=0;_&V.max_=0;
               &V.count=0 ; &V.scount=0 ; &V.mean=. ; &V.min=. ; &V.max=. ;
               %if "&event." ne "" %then %do ;
                   &V.events=0;
               %end ;
               %else %do ;
                   &V.smean=. ;&V.smin=.;&V.smax= .;&V.rmean=. ;&V.rmin=.;&V.rmax= .;&V.rrmean=. ;
               %end ;
           %end;

           do i=1 to &maxbins. ;
                count = c_tmp_[i] ;
                scount= sc_tmp_[i] ;
                %let vcountString  = ;
                %if "&PartitionVar" ne "" %then %do;
                    vcount = Vc_tmp_[i] ;
                    vscount= Vsc_tmp_[i] ;
                    %let vcountString = or vcount or vscount;
                %end;

                if scount or count &vcountString then do ;* write a record to microbins ;
                   %do i=1 %to &part;
                       %let V = ;
                       %if &i = 2 %then %let V = V;
                       if i gt 1 and &V.count ge 1 then do ;
                            &V.mean= &V.p_tmp_[i] / &V.count ;
                            &V.min= &V.m_tmp_[i] ;
                            &V.max= &V.x_tmp_[i] ;
                       end ;
                       %if "&event." ne "" %then %do;
                           &V.events= &V.e_tmp_[i] ;
                       %end ;
                       %else %do ;
                           if ^missing(&V.rr_tmp_[i]) then _&V.sse_  = _&V.sse_ + &V.rr_tmp_[i];
                           if &V.scount ge 1 then do ;
                              &V.smean= &V.sp_tmp_[i]/&V.scount ;
                              &V.smin=  &V.sm_tmp_[i] ;&V.smax=  &V.sx_tmp_[i] ;
                              &V.rmean= &V.rp_tmp_[i]/&V.scount ;
                              &V.rmin=  &V.rm_tmp_[i] ;&V.rmax=  &V.rx_tmp_[i] ;
                              &V.rrmean=&V.rr_tmp_[i]/&V.scount ;
                           end ;
                       %end ;
                   %end;
                   output microbins;
                end;
            end ;
            output  _newassess_stats;
          end ;
    run ;
    %if %sysfunc(exist(_newassess_stats)) %then %do;
       data _newassess_stats;
            set _newassess_stats;
            %if "&event" eq "" %then %do;
                 keep _sse_ _div_ _ase_ _rase_ _MEANP_ _STDP_ _CVP_
                 %if "&partitionVar" ne "" %then %do;
                     _Vsse_ _Vdiv_ _Vase_ _Vrase_ _VMEANP_ _vSTDP_ _VCVP_
                 %end;

                 ;

                 %do i=1 %to &part;
                    %let V = ;
                    %if &i = 2 %then %let V = V;

                     _&V.ASE_=.;_&V.RASE_=.;
                     if _&V.DIV_ ne 0 then do;
                        _&V.ASE_  = _&V.SSE_ / _&V.DIV_;
                        _&V.RASE_ = sqrt(_&V.ASE_);
                     end;
                     %if &i=1 %then %do;
                         _MEANP_ = &mean.;
                     _STDP_  = &std.;
                     %end;
                     %else %do;
                         _VMEANP_ = &vmean.;
                     _VSTDP_  = &vstd.;

                     %end;
                     if (_&V.MEANP_ ne . and _&V.MEANP_ ne 0) then _&V.CVP_   = _&V.STDP_/_&V.MEANP_;
                %end;
            %end;
            %else %do;
                %do i=1 %to &part;
                  %let V = ;
                  %if &i = 2 %then %let V = V;

                     keep _&V.misc_;
                     if (_&V.nobs_ ne 0 and _&V.nobs_ ne .)  then _&V.misc_ = _&V.misc_ / _&V.nobs_;
                     else _&V.misc_ = .;
                %end;
            %end;
       run;
   %end;

%mend aa_microbins_ds1_;
/*====================================================================================*/
/* %aa_isHP                                                                           */
/* Determines if a dataset is on a HP appliance                                       */
/*====================================================================================*/
%macro aa_isHP(data=, isgrid=);

    %let ENGINE = SAS;
    %if "&data" ne "" %then %do;
        %let dsid = %sysfunc(open(&DATA));
        %if &dsid> 0 %then %do;
            %let ENGINE = %upcase(%sysfunc(attrc(&dsid, ENGINE)));
        %end;
        %if &dsid > 0 %then %let dsid = %sysfunc(close(&dsid));
    %end;

    %let hpa_enabled = 0;

     %if "%scan(&data, 2, .)" = "" %then
         %let lib= work;
     %else
         %let lib = %scan(&data, 1, .);

    %let license = %sysfunc(sysprod(PRODNUM677));

    %if &license and (&engine = TERADATA or &engine = GREENPLM)  %then %do;
        %let prefix = %substr(%scan(%sysfunc(ranuni(0)), 2, .), 1, 6);
        %let _hpatestData= &lib..hpatest_&prefix;
        data &_hpatestData;
           length test 8;
           stop;
        run;

        %if %sysfunc(exist(&_hpatestdata)) %then %do;
            proc hpatest data=&_hpatestdata;
               performance nodes=all;
            run;
            proc delete data=&_hpatestdata;run;
        %end;
        %else %do;
            proc hpatest;
               performance nodes=all;
            run;
        %end;
    %end;

    %if "&isGrid" ne "" %then %do;
        %let &isGrid = 0;
        %if (&engine = TERADATA or &engine = GREENPLM) and &hpa_enabled %then %do;
            %let &isGrid = 1;
        %end;
        %put >>> ISGRID: &isGrid;
    %end;
%mend aa_isHP;

/*====================================================================================*/
/* %aa_model_register                                                                 */
/* Create bins to summarize a numeric variable without a sort                         */
/*====================================================================================*/
%macro aa_model_eval(
    /* required info -----------------------------------------------------------*/
    data    =,                          /* input data ;*/
    target  =,                          /* target variable ;*/
    var     =,                          /* numeric variable for ranking ;*/
    level   =,                          /* class or interval                    */
    event   =,                          /* event string (not variable name);*/
    /* output tables -----------------------------------------------------------*/
    out     =bins,                      /* output bin data name                 */
    binstats=binstats,                  /* output bin statistics like ks....    */
    /* optional general --------------------------------------------------------*/
    bins    =20,                        /* final number bins;*/
    fuzz    =0,                         /* handle ties ;*/
    freq    =,                          /* data freq ;*/
    /* optional create pseudo output for input to EM model import --------------*/
    expand  =expand,                    /* output data                          */
    cutoff  =,                          /* cutoff value for new classification  */
    /* optional pre-summary saves one pass of data -----------------------------*/
    min     =,
    max     =,
    hpds2   =,  /* Indicates if the microbins should be build using PROC HPDS2 */
    partitionvar=
    );
    %local mname ;%let mname= (aa_model_register) ;
    * ;
    %if "&data" eq "" %then
        %let data=train_scores ;
    %if not (%sysfunc(exist(&data)) or %sysfunc(exist(&data, VIEW))) %then %do;
          data _null_;
             putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_nodata_error, NOQUOTE,&data))";
          run;
        %goto exit ;
    %end ;
    * ;
    %if "&target" eq "" and %symexist(em_keytargetname) %then
        %let target =&em_keytargetname. ;
    %if "&target" eq "" %then %do ;
          data _null_;
               putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_notarget_error, NOQUOTE))";
          run;
        %goto exit ;
    %end ;
    * ;
    %if "&var" eq "" and %symexist(em_keytargetoutput) %then
        %let var =&em_keytargetoutput. ;
    %if "&var" eq "" %then %do ;
          data _null_;
             putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_nooutputvar_error, NOQUOTE))";
          run;
        %goto exit ;
    %end ;
    * ;

    %if "&level" eq "" and "&event" ne "" %then %do ;
        %let level=CLASS ;
    %end ;
    %if "&level" eq "" and %symexist(em_keytargetlevel) %then
        %let level=&em_keytargetlevel. ;
    %if "&level" eq "" %then %do ;
          data _null_;
             putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_notargetlevel_error, NOQUOTE))";
          run;
        %goto exit ;
    %end ;
    * ;

    %if %eval(&bins-2<0) %then %do;
        data _null_;
           putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_invalidnumbins_error, NOQUOTE))";
        run;
        %goto exit ;
    %end;

    %let level= %upcase(&level) ;
    %local imode ;%let imode=0 ;
    %if ("&level" eq "INTERVAL") or ("&level" eq "INT")  %then %do ;
        %let EVENT=;
        %let imode=1 ;
    %end;
    %else %do ;
        %let level = CLASS;
        %if "&event" eq "" and %symexist(em_keytargetevent) %then
            %let event=&em_keytargetevent. ;
        %if "&event" eq "" %then %do ;
                data _null_;
                   putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_notargetevent_error, NOQUOTE))";
               run;
            %goto exit ;
        %end ;
    %end ;
    * ;
    %put NOTE: &mname ;
    %put NOTE: data         =   &data. ;
    %put NOTE: target       =   &target. ;
    %put NOTE: level        =   &level. ;
    %put NOTE: var          =   &var.  ;
    %put NOTE: event        =   &event. ;
    %put NOTE: freq         =   &freq. ;
    %put NOTE: cutoff       =   &cutoff. ;
    %put NOTE: partitionVar =   &partitionVar. ;

    /*----------- check input data for vars ----------*/
    proc contents data=&data noprint out=_aamodelContents(keep=name type formatl formatd format);run ;
    %local novar noclass nofreq nofrom nointo format type;
    %let novar=1 ;%let noclass=1 ;%let nofreq=1 ;%let nofrom=1;%let nointo=1 ;%let format=;%let type=;

    %let deqevent =;
    data _null_ ;set _aamodelContents;
        length format $32 ;
        if upcase(name) eq upcase("&var.")     then do ;call symput("novar"   ,"0")  ;end ;
        if upcase(name) eq upcase("&freq.")    then do ;call symput("nofreq"  ,"0")  ;end ;
        if upcase(name) eq upcase("&target.")  then do ;
           call symput("type",  strip(put(type, best.)) ) ;
           call symput("notarget","0")  ;
           if format ne '' then do ;
              if formatl ne 0 then
                 format = strip(format)!!strip(put(formatl, best12.))!!'.'!!strip(put(formatd, best12.));
              if not index(format,'.') then format=strip(format) !! '.' ;
        end;

           call symput("format",strip(format)) ;
        end ;
        event = symget('EVENT');
        event = tranwrd(strip(event), "'", "''");
        call symput('DEQEVENT', strip(event));

    run ;
    proc delete data=_aamodelContents; run;

    %if "&novar." eq "1" %then %do ;
          data _null_;
             putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_rankingvarnotfound_error, NOQUOTE, &var))";
          run;
        %goto exit ;
    %end;
    %if "&notarget." eq "1" %then %do ;
          data _null_;
             putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_targetvarnotfound_error, NOQUOTE, &var))";
          run;
        %goto exit ;
    %end ;
    %if "&freq" ne "" and "&nofreq." eq "1" %then %do ;
          data _null_;
             putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_freqvarnotfound_error, NOQUOTE, &var))";
          run;
        %let freq= ;
    %end ;

    /*--- set bin and summary vars ---*/
    %local microsum microbin ;
    %if &imode %then %do ;
        %let microsum=&var. ;
        %let microbin=&target. ;
    %end;%else %do ;
        %let microsum=&target. ;
        %let microbin=&var. ;
    %end ;

    %put >>> HPDS2: &hpds2;

    %if "&hpds2" eq "" %then %do;
        %let _data = &data;
        %aa_isHP(data=&_data, isgrid=hpds2);

        %put >>> HPDS2: &hpds2;
    %end;
    %let _hpds2 = &hpds2;
    %if "&hpds2" = "Y" %then %let _hpds2 = 1;
    %else
    %if "&hpds2" = "N" %then %let _hpds2 = 0;

    /*===================================================*/
    /*--- MAP STEP -- create micro range bins -----------*/
    /*===================================================*/
    %local maxbins ;
    %let maxbins= 400;/*%sysfunc(max(100,min(20*&bins.,500))) ;*/
    %if ^&_hpds2 %then %do ;
        %aa_microbins_ds1_(&data.,&freq.,&microbin.,&microsum.,               &event.,                  &format., &type., &maxbins.,&fuzz.,&min.,&max., &partitionvar.);
    %end ;%else %do ;
        %aa_microbins_hpds2_(&data.,&freq.,&microbin.,&microsum.,%nrbquote(&event),%nrbquote(&deqevent),&format., &type., &maxbins.,&fuzz.,&min.,&max., &partitionvar.);
    %end ;
    /*===================================================*/
    /*--- START REDUCE STEP -- merge bins into centiles -*/
    /*===================================================*/
    /*--- sort microbins merged from all captains -------*/
    /*--- first get counts and sums ---------------------*/

    %let num=1;
    %let prefix =;
    %if "&partitionVar" ne "" %then %do;
        %let num = 2;
    %end;

%do i= 1 %to &num;
        %let V=;
        %if "&i" = "2" %then
            %let V = V;

    proc sort data=microbins(drop=
       %if "&event" eq "" %then %do ;
           _&V.sse_ _&V.div_
       %end;
       %else %do;
           _&V.misc_ _&V.nobs_
       %end;
        )
       ;
       by descending &V.mean ;
    run ;

    %local sumbins sumobs sumcount misscount sumscount missscount sumevents eventmiss ;

    data _null_ ;set microbins end=eof;

        retain sumcount misscount sumscount misscount 0;
        %if &imode %then %do ;retain sumevents eventmiss 0;%end ;

        if missing(&V.mean) then do ;
            misscount+&V.count ;
            %if not &imode %then %do;eventmiss+&V.events ;%end ;
        end ;
        else do ;
            sumcount+&V.count ;
            sumscount+&V.scount ;
            %if not &imode %then %do;sumevents+&V.events ;%end ;
        end ;
        if eof then do ;
            missscount= sumcount - sumscount ;
            call symput("SUMBINS"   , strip(put(_N_,best.))) ;
            call symput("SUMCOUNT"  , strip(put(sumcount,best.)));
            call symput("MISSCOUNT" , strip(put(misscount,best.))) ;
            call symput("SUMSCOUNT" , strip(put(sumscount,best.))) ;
            call symput("MISSSCOUNT", strip(put(missscount,best.))) ;
            call symput("SUMOBS"    , strip(put(sumcount+misscount,best.))) ;
            %if not &imode %then %do ;
            call symput("SUMEVENTS" , put(sumevents,best.)) ;
            call symput("EVENTMISS" , strip(put(eventmiss,best.)) ) ;
            %end ;
        end ;
    run ;
    /*--- merge microbins by slicing ---*/
    %if &imode %then %do ;
        data _aamodel_tempbins ;set microbins(where=(&V.mean ne .)) end=eof ;
           %if "&partitionVar" ne "" %then %do;
               keep DataRole;
               length DataRole $8;
               %if &i =1 %then %do;
                 retain DataRole 'TRAIN';
             %end;
             %else %do;
                 retain DataRole 'VALIDATE';
             %end;
           %end;

            keep count target_mean target_min target_max ;
            keep predicted_count predicted_mean predicted_min predicted_max ;
            keep residual_mean residual_min residual_max residual_squared_mean ;

            retain binsize 0 vc 0 vb 0 ;
            retain target_mean 0 target_min 1e16 target_max -1e16 ;
            retain predicted_count 0 predicted_mean 0 predicted_min 1e16 predicted_max -1e16 ;
            retain residual_mean 0 residual_min 1e16 residual_max -1e16  residual_squared_mean 0 ;
            if _n_ eq 1 then binsize= int(&sumcount./&bins.+0.5) ;
            *;
            if &V.min  ne . then target_min    = min(target_min,       &V.min) ;
            if &V.max  ne . then target_max    = max(target_max,       &V.max) ;
            if &V.smin ne . then predicted_min = min(predicted_min,    &V.smin) ;
            if &V.smax ne . then predicted_max = max(predicted_max,    &V.smax) ;
            if &V.rmin ne . then residual_min  = min(residual_min,     &V.rmin) ;
            if &V.rmax ne . then residual_max  = max(residual_max,     &V.rmax) ;
            *;
            imax=&V.count ;
            snum=0 ;
            do i= 1 to imax ;
                * target counts ;
                vc= vc + 1 ;
                if &V.mean ne .  then do ;
                    target_mean= target_mean + &V.mean  ;
                end ;
                * predicted counts ;
                snum+1 ;
                if snum le scount then do ;
                    predicted_count= predicted_count + 1 ;
                    predicted_mean=  predicted_mean  + &V.smean ;
                    residual_mean= residual_mean + &V.rmean ;
                    residual_squared_mean + &V.rrmean ;
                end ;
                * slice bin and output if maxbinsize reached ;
                if vc eq binsize and vb lt (&bins.-1) then do ;
                    vb+1 ;count=vc ;
                    if count gt 0 then do ;
                        target_mean= target_mean / count ;
                    end ;
                    if predicted_count gt 0 then do ;
                        predicted_mean= predicted_mean / predicted_count ;
                        residual_mean= residual_mean / predicted_count ;
                        residual_squared_mean= residual_squared_mean / predicted_count ;
                    end ;
                    else do ;
                        predicted_mean= . ;predicted_max=. ;predicted_min= . ;
                        residual_mean= . ;residual_max=. ;residual_min= . ;
                        residual_squared_mean= . ;
                    end ;
                    output ;
                    vc=0 ;
                    predicted_count= 0 ;
                    target_mean=0 ;target_min= &V.min ;target_max=&V.max ;
                    predicted_mean=0 ;predicted_min= &V.smin;predicted_max= &V.smax ;
                    residual_mean=0  ;residual_min = &V.rmin;residual_max = &V.rmax ;
                    residual_squared_mean=0 ;
                end ;
            end ;
            if eof then do ;
                vb+1 ;count=vc ;
                if count gt 0 then do ;
                    target_mean= target_mean / count ;
                end ;
                if predicted_count gt 0 then do ;
                    predicted_mean= predicted_mean / predicted_count ;
                    residual_mean= residual_mean / predicted_count ;
                    residual_squared_mean= residual_squared_mean / predicted_count ;
                end ;
                else do ;
                    predicted_mean= . ;predicted_max=. ;predicted_min= . ;
                    residual_mean= . ;residual_max=. ;residual_min= . ;
                    residual_squared_mean= . ;
                end ;
                output ;
            end ;
        run ;
    %end;
    %else %do  ;
        data _aamodel_tempbins ;set microbins(where=(&V.mean ne .)) end=eof ;
           %if "&partitionVar" ne "" %then %do;
               keep DataRole;
             length DataRole $8;
             %if &i =1 %then %do;
                 retain DataRole 'TRAIN';
             %end;
             %else %do;
                 retain DataRole 'VALIDATE';
             %end;
         %end;

            keep count events predicted_mean predicted_min predicted_max ;
            length count events  8 ;
            retain binsize 0 vc 0 ve 0 vb 0 vp 0 vec 0 vx -1e16 vm 1e16 ;
            if _n_ eq 1 then
                        binsize= int(&sumcount./&bins.+0.5) ;
            if &V.count then do;
               rate= &V.events/&V.count ;imax= &V.count ;
               do i= 1 to imax ;
                  vc= vc + 1 ;ve= ve + rate ;vp= vp + &V.mean ;
                  vm= min(vm,&V.min) ;vx= max(vx,&V.max) ;
                  if vc eq binsize and vb lt (&bins.-1) then do ;
                      ve= int(ve+0.5);
                      if (vec + ve) > &sumevents. then ve= (&sumevents.-vec) ;
                      vec= ve + vec ;
                      vb+1 ;count=vc ;events=ve  ;
                              if vc then do ;
                                    predicted_mean= vp/vc;
                                    predicted_min = vm;
                                    predicted_max = vx ;
                              end ;
                      output ;
                      vc=0 ;ve=0 ;vp= 0 ;vx= -1e16 ;vm= 1e16 ;
                  end;
               end;
            end;
            if eof then do ;
                ve= max(0,&sumevents. - vec) ;
                vb+1 ;count= vc ;events=ve ;
                        if vc then do ;
                              predicted_mean= vp/vc;
                              predicted_min=vm;
                              predicted_max=vx ;
                        end ;
                output ;
            end ;
         run;
    %end;
    /*===================================================*/
    /*--- end of REDUCE STEP ----------------------------*/
    /*===================================================*/
    /*--- Build derived values ---*/
    %if &imode %then %do ;
        data _aamodel_tempbins;
           length Target $32;
           Label
             Target         = "%sysfunc(sasmsg(sashelp.aacomp, model_target_vlabel, NOQUOTE))"
             %if "&partitionVar" ne "" %then %do;
                 DataRole       = "%sysfunc(sasmsg(sashelp.aacomp, model_datarole_vlabel, NOQUOTE))"
             %end;

            depth          = "%sysfunc(sasmsg(sashelp.aacomp, model_depth_vlabel, NOQUOTE))"
            count          = "%sysfunc(sasmsg(sashelp.aacomp, model_nobs_vlabel, NOQUOTE))"
            c_count        = "%sysfunc(sasmsg(sashelp.aacomp, model_cumnobs_vlabel, NOQUOTE))"
            predicted_count= "%sysfunc(sasmsg(sashelp.aacomp, model_predcount_vlabel, NOQUOTE))"
            c_predicted_count= "%sysfunc(sasmsg(sashelp.aacomp, model_cumpredcount_vlabel, NOQUOTE))"
                  target_mean = "%sysfunc(sasmsg(sashelp.aacomp, model_targetbinMean_vlabel, NOQUOTE))"
            target_max  = "%sysfunc(sasmsg(sashelp.aacomp, model_targetmax_vlabel, NOQUOTE))"
            target_min  = "%sysfunc(sasmsg(sashelp.aacomp, model_targetmin_vlabel, NOQUOTE))"
            predicted_mean = "%sysfunc(sasmsg(sashelp.aacomp, model_predictedbinMean_vlabel, NOQUOTE))"
            predicted_max  = "%sysfunc(sasmsg(sashelp.aacomp, model_predictedmax_vlabel, NOQUOTE))"
            predicted_min  = "%sysfunc(sasmsg(sashelp.aacomp, model_predictedmin_vlabel, NOQUOTE))"
            residual_mean  = "%sysfunc(sasmsg(sashelp.aacomp, model_residualbinMean_vlabel, NOQUOTE))"
            residual_max   = "%sysfunc(sasmsg(sashelp.aacomp, model_residualmax_vlabel, NOQUOTE))"
            residual_min   = "%sysfunc(sasmsg(sashelp.aacomp, model_residualmin_vlabel, NOQUOTE))"
            residual_squared_mean = "%sysfunc(sasmsg(sashelp.aacomp, model_residualsquaredbinmean_vlabel, NOQUOTE))"
            ;
            set _aamodel_tempbins  end=eof ;
            retain target "&target" depth 0 count 0 c_count 0 predicted_count 0 c_predicted_count 0 ;
            c_count = c_count + count ;
            if predicted_count ne . then c_predicted_count= c_predicted_count + predicted_count ;
            depth = 100 * c_count / &sumcount. ;
            if eof then do ;
                call symput( "PREDCOUNT", put(c_predicted_count,best.)) ;
                call symput( "PREDMISS" , put(c_count-c_predicted_count,best.)) ;
            end ;
        run ;
    %end ;
    %else %do ;
        %put NOTE: &mname total count= &sumcount., total events= &sumevents., bins= &sumbins. ;
        /*--- save the ks and classification rate values -----*/
        %local  ks ksref ksdepth kscut ksr ;
        %local  cr crdepth crcut mcut mdepth ;
        /*--- compute cumulative stats above bin threshold ---*/
        %local sumnonevents ;
        %let sumnonevents= %eval(&sumcount. - &sumevents.) ;
        * ;
        %let KS=.;
        %let KSREF=.;
        %let KSR=.;
        %let KSDEPTH=.;
        %let KSCUT=.;
        %let CR=.;
        %let CRDEPTH=.;
        %let CRCUT=.;
        %let MDEPTH=.;
        %let MCUT=.;

        data _aamodel_tempbins ;set _aamodel_tempbins  end=eof ;
           retain depth 0 count 0 c_count 0 events 0 c_events 0 c_nonevents 0 ;
             c_count = c_count + count ;
              depth = 100 * c_count / &sumcount. ;
              c_events= c_events + events ;
              c_event_pct= 100*c_events/&sumevents. ;
              nonevents= count-events ;
              c_nonevents= c_nonevents + nonevents ;
              c_nonevent_pct=100*c_nonevents/&sumnonevents. ;
              *;
              if count gt 0 then correct_rate= 100*( c_events + &sumnonevents. - c_nonevents ) / &sumcount. ;
              if count gt 0 then error_rate= 100 - correct_rate ;
              * event rates ;
              if count gt 0 then event_rate= 100 * events / count ;
              if count gt 0 then nonevent_rate= 100 * nonevents / count ;
              separation= c_event_pct - c_nonevent_pct ;
              * gain and lift ;
              if count gt 0 then do;
                 lift  = (events/count) / (&sumevents./&sumcount.) ;
                 gain = 100*(lift - 1);
              end;
              if count gt 0 then do;
                 c_lift= (c_events/c_count) / (&sumevents./&sumcount.);
                 c_gain = 100*(c_lift -1);
              end;
              * ;
              sensitivity = c_events / &sumevents. ;
              specificity= (&sumnonevents. - c_nonevents) / &sumnonevents. ;
              one_minus_specificity=1-specificity ;
              *;
              * get KS and cutoff type measures ;
              drop ks ksr ksdepth kscut cr crdepth crcut mcut mdepth ksref ;
                  retain ks 0 ksr 0 ksref 0 ksdepth 0 kscut 0 cr 0 crdepth 0 crcut 0 mcut 0 mdepth 0 ;
                  if separation > ksr then do ;
                        ksr= separation ;
                        ks= sensitivity-one_minus_specificity;
                        ksref= one_minus_specificity ;
                        ksdepth= depth ;kscut= predicted_mean ;
                  end ;
                  if c_events < 0.51*&sumevents. then do ;mcut=predicted_mean ;mdepth=depth ;end ;
              if correct_rate > cr then do ;cr= correct_rate ;crdepth= depth;crcut= predicted_mean ;end ;
                  * ;
              /*--- Add other measures here ------------------------------*/
              /* sensitivity, specificity, etc..............              */
              if eof then do ;
                  call symput( "KS"          , strip(put(ks,best.))) ;
                  call symput( "KSREF"       , strip(put(ksref,best.))) ;
                  call symput( "KSR"         , strip(put(ksr,best.))) ;
                  call symput( "KSDEPTH"     , strip(put(ksdepth,best.))) ;
                  call symput( "KSCUT"       , strip(put(kscut,best.))) ;
                  call symput( "CR"          , strip(put(cr,best.))) ;
                  call symput( "CRDEPTH"     , strip(put(crdepth,best.))) ;
                  call symput( "CRCUT"       , strip(put(crcut,best.))) ;
                  call symput( "MDEPTH"      , strip(put(mdepth,best.))) ;
                  call symput( "MCUT"        , strip(put(mcut,best.))) ;
              end ;
        run ;
        * add zero row ;
        data _aamodelZero;
           %if "&partitionVar" ne "" %then %do;
               length DataRole $8;
               %if &i =1 %then %do;
                   retain DataRole 'TRAIN';
               %end;
               %else %do;
                   retain DataRole 'VALIDATE';
               %end;
           %end;
            depth=0 ;count=0 ;c_count=0 ;
            events=0 ;c_events= 0 ;nonevents= 0 ;c_nonevents=0 ;
            c_event_pct= 0;c_nonevent_pct=0 ;
            sensitivity=0 ;specificity=0 ;one_minus_specificity=0 ;
            output ;
        run ;
        * Combine ;
        data _aamodel_tempbins;
           length Target $32;
           label
             Target         = "%sysfunc(sasmsg(sashelp.aacomp, model_target_vlabel, NOQUOTE))"
             %if "&partitionVar" ne "" %then %do;
                 DataRole       = "%sysfunc(sasmsg(sashelp.dmine, rpt_datarole_vlabel, NOQUOTE))"
             %end;
             depth          = "%sysfunc(sasmsg(sashelp.aacomp, model_depth_vlabel, NOQUOTE))"
             count          = "%sysfunc(sasmsg(sashelp.aacomp, model_nobs_vlabel, NOQUOTE))"
             c_count        = "%sysfunc(sasmsg(sashelp.aacomp, model_cumnobs_vlabel, NOQUOTE))"
                   events         = "%sysfunc(sasmsg(sashelp.aacomp, model_eventcount_vlabel, NOQUOTE))"
             nonevents      = "%sysfunc(sasmsg(sashelp.aacomp, model_noneventcount_vlabel, NOQUOTE))"
             c_events       = "%sysfunc(sasmsg(sashelp.aacomp, model_cumeventcount_vlabel, NOQUOTE))"
             c_nonevents    = "%sysfunc(sasmsg(sashelp.aacomp, model_cumnoneventcount_vlabel, NOQUOTE))"
             c_event_pct    = "%sysfunc(sasmsg(sashelp.aacomp, model_cumpctevent_vlabel, NOQUOTE))"
             c_nonevent_pct = "%sysfunc(sasmsg(sashelp.aacomp, model_cumpctnonevent_vlabel, NOQUOTE))"
             correct_rate   = "%sysfunc(sasmsg(sashelp.aacomp, model_correctrate_vlabel, NOQUOTE))"
             error_rate     = "%sysfunc(sasmsg(sashelp.aacomp, model_errorrate_vlabel, NOQUOTE))"
             event_rate     = "%sysfunc(sasmsg(sashelp.aacomp, model_eventrate_vlabel, NOQUOTE))"
             nonevent_rate  = "%sysfunc(sasmsg(sashelp.aacomp, model_noneventrate_vlabel, NOQUOTE))"
             sensitivity    = "%sysfunc(sasmsg(sashelp.aacomp, model_sensitivity_vlabel , NOQUOTE))"
             specificity    = "%sysfunc(sasmsg(sashelp.aacomp, model_specificity_vlabel, NOQUOTE))"
             predicted_mean = "%sysfunc(sasmsg(sashelp.aacomp, model_predictedbinMean_vlabel, NOQUOTE))"
             predicted_max  = "%sysfunc(sasmsg(sashelp.aacomp, model_predictedmax_vlabel, NOQUOTE))"
             predicted_min  = "%sysfunc(sasmsg(sashelp.aacomp, model_predictedmin_vlabel, NOQUOTE))"
             one_minus_specificity = "%sysfunc(sasmsg(sashelp.aacomp, model_oneminusspecificity_vlabel , NOQUOTE))"
             separation     = "%sysfunc(sasmsg(sashelp.aacomp, model_ksr_vlabel, NOQUOTE))"
             lift           = "%sysfunc(sasmsg(sashelp.aacomp, model_lift_vlabel, NOQUOTE))"
             c_lift         = "%sysfunc(sasmsg(sashelp.aacomp, model_cumlift_vlabel, NOQUOTE))"
             gain           = "%sysfunc(sasmsg(sashelp.aacomp, model_gain_vlabel, NOQUOTE))"
             c_gain         = "%sysfunc(sasmsg(sashelp.aacomp, model_c_gain_vlabel, NOQUOTE))";
             retain Target "&target";
             set _aamodelZero _aamodel_tempbins;
         run ;
        proc delete data= _aamodelZero; run;
    %end ;
    /*--- create bin statistics table of point measures ---*/
    %let lv= t;
    %if &i=2 %then
        %let lv = v;

    data _aa_model_tempbinstats ;
        length Target $32 Level $32;
        LABEL
                Target = "%sysfunc(sasmsg(sashelp.aacomp, model_target_vlabel, NOQUOTE))"
            Level  = "%sysfunc(sasmsg(sashelp.aacomp, model_targetlevel_vlabel, NOQUOTE))"
            Var    = "%sysfunc(sasmsg(sashelp.aacomp, model_outputvar_vlabel, NOQUOTE))"
            NBins  = "%sysfunc(sasmsg(sashelp.aacomp, model_numbins_vlabel, NOQUOTE))"

            &v.NObs   = "%sysfunc(sasmsg(sashelp.aacomp, model_&lv.nobs_vlabel, NOQUOTE))"
            &v.TargetCount = "%sysfunc(sasmsg(sashelp.aacomp, model_&lv.targetcount_vlabel, NOQUOTE))"
            &v.TargetMiss  = "%sysfunc(sasmsg(sashelp.aacomp, model_&lv.targetmiss_vlabel, NOQUOTE))"
            &v.PredCount   = "%sysfunc(sasmsg(sashelp.aacomp, model_&lv.predcount_vlabel, NOQUOTE))"
            &v.PredMiss    = "%sysfunc(sasmsg(sashelp.aacomp, model_&lv.predmiss_vlabel, NOQUOTE))"
            %if not &imode %then %do ;
                &v.Event       =   "%sysfunc(sasmsg(sashelp.aacomp, model_targetevent_vlabel, NOQUOTE))"
                &v.EventCount  =   "%sysfunc(sasmsg(sashelp.aacomp, model_&lv.eventcount_vlabel, NOQUOTE))"
                &v.NonEventCount =   "%sysfunc(sasmsg(sashelp.aacomp, model_&lv.noneventcount_vlabel, NOQUOTE))"
                &v.EventMiss   =   "%sysfunc(sasmsg(sashelp.aacomp, model_eventmiss_vlabel, NOQUOTE))"
                &v.KSR         =   "%sysfunc(sasmsg(sashelp.aacomp, model_&lv.ksr_vlabel, NOQUOTE))"
                &v.KS          =   "%sysfunc(sasmsg(sashelp.aacomp, model_&lv.ks_vlabel, NOQUOTE))"
                &v.KSDepth     =   "%sysfunc(sasmsg(sashelp.aacomp, model_&lv.ksdepth_vlabel, NOQUOTE))"
                &v.KSCut       =   "%sysfunc(sasmsg(sashelp.aacomp, model_&lv.kscut_vlabel, NOQUOTE))"
                &v.KSRef       =   "%sysfunc(sasmsg(sashelp.aacomp, model_&lv.ksref_vlabel, NOQUOTE))"
                &v.CR          =   "%sysfunc(sasmsg(sashelp.aacomp, model_&lv.cr_vlabel, NOQUOTE))"
                &v.CRDepth     =   "%sysfunc(sasmsg(sashelp.aacomp, model_&lv.crdepth_vlabel, NOQUOTE))"
                &v.CRCut       =   "%sysfunc(sasmsg(sashelp.aacomp, model_&lv.crcut_vlabel, NOQUOTE))"
                &v.MDepth      =   "%sysfunc(sasmsg(sashelp.aacomp, model_&lv.mdepth_vlabel, NOQUOTE))"
                &v.MCut        =   "%sysfunc(sasmsg(sashelp.aacomp, model_&lv.mcut_vlabel, NOQUOTE))"
            %end ;
            ;
            Target          =   "&TARGET"   ;
            Level           =   "&LEVEL"    ;
            Var             =   "&VAR"      ;
            NBins           =   &BINS.      ;
            &v.NObs            =   &SUMOBS.    ;
            %if &imode %then %do  ;
                &v.TargetCount =   &SUMCOUNT.  ;/* this was the bin var */
                &v.TargetMiss  =   &MISSCOUNT. ;
                &v.PredCount   =   &SUMSCOUNT. ;/* this was the sum var */
                &v.PredMiss    =   &MISSSCOUNT.;
            %end ;
            %else %do ;
                Event       =   "&EVENT"    ;
                &v.TargetCount =   &SUMSCOUNT. ;/* this was the sumvar */
                &v.TargetMiss  =   &MISSSCOUNT.;
                &v.PredCount   =   &SUMCOUNT.  ;/* this was the binvar */
                &v.PredMiss    =   &MISSCOUNT. ;
                &v.EventCount  =   &SUMEVENTS. ;
                &v.NonEventCount =   &SUMCOUNT. - &SUMEVENTS. ;
                &v.EventMiss   =   &eventmiss. ;
                &v.KSR         =   &KSR.       ;
                &v.KS          =   &KS.        ;
                &v.KSDepth     =   &KSDEPTH.   ;
                &v.KSCut       =   &KSCUT.     ;
                &v.KSRef       =   &KSREF.     ;
                &v.CR          =   &CR.        ;
                &v.CRDepth     =   &CRDEPTH.   ;
                &v.CRCut       =   &CRCUT.     ;
                &v.MDepth      =   &MDEPTH.    ;
                &v.MCut        =   &MCUT.      ;
            %end ;
            output ;
    run ;

        /*--- expand bins to create data for input to EM assess ----*/
        /*--- Can be used in the model import node -----------------*/
        %if not &imode %then %do ;
            %local cut ;
            %let cut= &cutoff ;
            %if "&cutoff." eq ""    %then %do;%let cut= &KSCUT  ;%end ;
            %if "&cutoff." eq "KS"  %then %do;%let cut= &KSCUT  ;%end ;
            %if "&cutoff." eq "CR"  %then %do;%let cut= &CRCUT  ;%end ;
            %if "&cutoff." eq "MED" %then %do;%let cut= &MCUT   ;%end ;
            *;
                data _null_ ;
                      call symput('cut',strip(put(&cut.,best6.2))) ;
                run ;
            data _aamodel_tempexpand;
               %if "&partitionVar" ne "" %then %do;
                   keep DataRole;
                   length DataRole $8;
                   %if "&i" ="1" %then %do;
                       retain DataRole 'TRAIN';
                   %end;
                   %else %do;
                       retain DataRole 'VALIDATE';
                   %end;
               %end;

               length target predict $32 label $200 ;
                keep target predict prob freq  ;
                label Target = "%sysfunc(sasmsg(sashelp.aacomp, model_targetvalue_vlabel, NOQUOTE))"
                      Predict= "%sysfunc(sasmsg(sashelp.aacomp, model_predictionatcutoff_vlabel, NOQUOTE, &cut))"  ;

                set _aamodel_tempbins(where=(predicted_mean ne .)) ;
                * cutoff values ;
                prob = predicted_mean ;
                if prob ge &cut. then
                    predict = "event" ;
                else
                predict = "non-event" ;
                * events ;
                freq= events ;
                target = "event" ;
                if freq then output ;
                * nonevents ;
                freq= count - events ;
                target = "non-event" ;
                if freq then output ;
            run ;

            proc append base= &expand. data=_aamodel_tempexpand force;
          run;
          proc delete data=_aamodel_tempexpand;run;

    %end ;

    proc append base=&out. data=_aamodel_tempbins force;
    run;
    proc delete data=_aamodel_tempbins;run;

    %if ^%sysfunc(exist(&binstats)) %then %do;
        data &binstats;
             set _aa_model_tempbinstats;
        run;
    %end;
    %else %do;
        data &binstats;
           merge &binstats _aa_model_tempbinstats;
        run;
   %end;
   proc delete data=_aa_model_tempbinstats;run;

%end; /* Partition Loop */

    %if %sysfunc(exist(_newassess_stats)) %then %do;
        data &binstats;
           merge &binStats _newassess_stats;
           label
            %if &imode %then %do ;
               _div_    = "%sysfunc(sasmsg(sashelp.aacomp, model_div_vlabel, NOQUOTE))"
               _sse_    = "%sysfunc(sasmsg(sashelp.aacomp, model_sse_vlabel, NOQUOTE))"
               _ase_    = "%sysfunc(sasmsg(sashelp.aacomp, model_ase_vlabel, NOQUOTE))"
               _rase_   = "%sysfunc(sasmsg(sashelp.aacomp, model_rase_vlabel, NOQUOTE))"
               _MEANP_  = "%sysfunc(sasmsg(sashelp.aacomp, model_meanp_vlabel, NOQUOTE))"
               _STDP_   = "%sysfunc(sasmsg(sashelp.aacomp, model_stdp_vlabel, NOQUOTE))"
               _CVP_    = "%sysfunc(sasmsg(sashelp.aacomp, model_cvp_vlabel, NOQUOTE))"

               %if "&partitionVar" ne "" %then %do;
                 _VDIV_    = "%sysfunc(sasmsg(sashelp.aacomp, model_vdiv_vlabel, NOQUOTE))"
                 _VSSE_    = "%sysfunc(sasmsg(sashelp.aacomp, model_vsse_vlabel, NOQUOTE))"
                 _VASE_    = "%sysfunc(sasmsg(sashelp.aacomp, model_vase_vlabel, NOQUOTE))"
                 _VRASE_   = "%sysfunc(sasmsg(sashelp.aacomp, model_vrase_vlabel, NOQUOTE))"
                 _VMEANP_   = "%sysfunc(sasmsg(sashelp.aacomp, model_vmeanp_vlabel, NOQUOTE))"
                 _VSTDP_    = "%sysfunc(sasmsg(sashelp.aacomp, model_vstdp_vlabel, NOQUOTE))"
                 _VCVP_     = "%sysfunc(sasmsg(sashelp.aacomp, model_vcvp_vlabel, NOQUOTE))"

             %end;
           %end;
           %else %do;
               _misc_   = "%sysfunc(sasmsg(sashelp.aacomp, model_misc_vlabel, NOQUOTE))"
               %if "&partitionVar" ne "" %then %do;
                  _VMISC_   = "%sysfunc(sasmsg(sashelp.aacomp, model_vmisc_vlabel, NOQUOTE))"
             %end;
            %end;
               ;
        run;
        proc delete data=_newassess_stats;
        run;
    %end;


    * reset some global macro vars ;
   %if "&em_dataset" eq "" %then %let em_dataset= &data ;
    %let em_keytargetname = &target ;
    %let em_keytargetevent= &event ;
    %let em_keytargetlevel= &level ;

%exit:
%mend aa_model_eval;



%put NOTE: %sysfunc(sasmsg(sashelp.aacomp, model_loadingregister_note , NOQUOTE));
/*====================================================================================*/
/* NAME:  aa_Model_Register                                                                    */
/* BY:    sasdzl                                                                      */
/* DATE:  Feb23, 2012                                                            */
/*------------------------------------------------------------------------------------*/
/* Registering Models to the Model Repository                                         */
/*====================================================================================*/
/* Global macro vars declared in one place                                            */
/*====================================================================================*/

/*=====================================================================================*/
/* NAME:  aa_itemScoreExtract                                                          */
/* BY:    jadean                                                                       */
/* DATE:  Feb 3, 2012                                                                  */
/*-------------------------------------------------------------------------------------*/
/* Extract information from an itemstore created by a PML procedure                    */
%macro aa_itemstoreExtract(
/* Required Information ---------------------------------------------------------------*/
   itemstore=,                                  /*             name of itemstore       */
/* Optional Information ---------------------------------------------------------------*/
   varDs=work._varlookup,    /*  dataset to contain the following                      */
         /*         _role =(interval, nominal, target, segment) variable role          */
         /*         descr =(_Model_Effects_, _Class, _Response_) variable type         */
         /*         value =the variable name                                           */
         /*         levels=the number of levels (missing for numeric variables)        */
         /*         values=a space seperated list of the levels                        */
   infoDs=work._modelinfo,                             /*   Model information          */
   pgmFile=,               /* file to the program that created the itemstore           */
   scoreCodeFile=,         /* file to extract the score code from the itemstore to     */
   lookup=,                /* LOOKUP algorithm for class levels                        */
   fitstats=work._fitStats, /* data set to contain the model fit statistics            */
   debug=N
);


    %if "&pgmFile" eq "" %then %do;
        %let pgmFile = temp;
    %end;
    filename _fpgm "&pgmFile" encoding="UTF-8" NOBOM;;

    /*  Call Proc PLM extract to get StoreInfo*/
    proc plm restore = &itemStore;
        show all;
        ods output StoreInfo=_ItemStoreInfo;
    run;

      %let _hasClass = 0;
      %if %sysfunc(exist(_ItemStoreInfo)) %then %do;
          data _null_;
               set _itemStoreInfo end=eof;
               where upcase(_Descr)='_CLASS_';
           if eof then call symput('_hasClass', '1');
       run;
    %end;
    %else %do;
        data _null_;
          putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_itemstoreinfonotfound_error, NOQUOTE))";
          call symput('aaModelErrorString', 'model_itemstoreinfonotfound_error');
        run;
        %let syscc = 2000;
        %goto endis;
    %end;

    /*  Call Proc PLM extract the needed information*/
    proc plm restore = &itemStore;
        show all;
        ods output variableInfo=_varinfo;
        ods output program=_program;
        ods output fitstatistics=&fitstats;
        %if &_hasClass %then %do;
           ods output ClassLevels=_ClassLevels;
       %end;
         %if "&scorecodeFile" ne "" %then %do;
             code file = "&scorecodeFile"
             %if "&lookup" ne "" %then %do;
                 lookup=&lookup
             %end;
             ;
         %end;
    run;

    /*  Convert _program data set to a program*/
    %if %sysfunc(exist(_program)) %then %do;
        data _null_;
           file _fpgm;
           set _program;
           put descr;
        run;
   %end;
   %else %do;
       data _null_;
          putlog "WARNING: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_itemstorenoprogram_warning, NOQUOTE))";
       run;
   %end;

   /*Get information from the itemstore*/
    data &infoDs;
       length Name $16 Value $32;
       keep Name Value;
        set _ItemStoreInfo;
        if _Descr ='_Input_Data_'   then do;Name='DATA';VALUE=value;output;end;
        if _Descr ='_Response_'     then do;Name='TARGET';VALUE=value;output;end;
        if _Descr ='_Distribution_' then do;Name='DISTRIBUTION';VALUE=value;output;end;
        if _Descr ='_Created_By_'   then do;Name='MININGALGORITHM';VALUE=value;output;end;
    run;

    %if %sysfunc(exist(_classlevels)) %then %do;
        /*Merge class level and var info tables*/
        %let _itemStoreError = 0;
        proc sort data=_classlevels;by class;run;
        proc sort data=_varinfo;by value;run;
        data _merge1;
           merge _varinfo(in=_in1) _classlevels(in=_in2 rename=(class=value));
           by value;
           /*      ERROR handling*/
           if _in2 and not _in1 then do;
              putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_inconsistentitemstore_error, NOQUOTE))";
              call symput('aaModelErrorString', 'model_inconsistentitemstore_error');
              call symput('_itemStoreError', '1');
              stop;
           end;
           if _in1;
        run;
        %if &_itemStoreError %then %do;
                  %let syscc = 2000;
            %goto endis;
        %end;

        data &varDs;
           length UNAME $32 AALevel $10 AARole $12.;
           set _merge1(where=(descr^="_Class_"));
           AARole='INPUT';
           if descr in("_Model_Effects_", "_Response_") then do;
              if levels=.      then AALevel="INTERVAL";
              else if levels>2 then AALevel="NOMINAL";
              else if levels=2 then AALevel="BINARY";
              else if levels=1 then AALevel="UNARY";
              else do;
                 call symput('_itemlevel', strip(put(levels, best.)));
                 call symput('_itemvar',    strip(value));
                 putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_wrongvarlevel_error, NOQUOTE, &_itemlevel, &_itemvar))";
                 call symput('aaModelErrorString', 'model_wrongvarlevel_error');
                 stop;
              end;
           end;
           if descr="_by_" then do;
              AARole="SEGMENT";
              AALEVEL = "NOMINAL";
           end;
           if descr="_Response_" then AARole="TARGET";
           UNAME = upcase(VALUE);
        run;
    %end;
      %else %do;
        data &varDs;
           length UNAME $32 AALevel $10 AARole $12.;
           retain AALevel "INTERVAL";
           set _varinfo;
           AARole='INPUT';
           if descr="_by_" then do;
              AARole="SEGMENT";
              AALEVEL = "NOMINAL";
           end;
           if descr="_Response_" then AARole="TARGET";
           UNAME = upcase(VALUE);
        run;
      %end;

    /*  clean up datasets*/
    %if &debug=N %then %do;
        proc datasets library=work nolist nowarn;
           delete _classlevels;
           delete _varinfo;
           delete _variable_type;
           delete _program;
           delete _merge1;
           delete _itemstoreinfo;
        run;
       quit;
    %end;
    filename _fpgm;

    %endis:

%mend aa_itemstoreExtract;


%GLOBAL  msdsep;
/*====================================================================================*/
/*====================================================================================*/
/* NAME:  aa_model_getVarMeta                                                         */
/* BY:    sasdzl                                                                      */
/* DATE:  January 10, 2012                                                            */
/*------------------------------------------------------------------------------------*/
/* Determine the basic variable metadata                                              */
/*====================================================================================*/
%macro aa_model_getVarMeta(
/* Required Information ---------------------------------------------------------------*/
data=,                                     /*                  input-training data     */
output=,                                    /*                  generated output table */
/* Optional Information ---------------------------------------------------------------*/
itemstoreVar=,                             /*     variable information from itemstore  */
type=,                                     /*                  type INPUT or OUTPUT    */
target=,                                   /*                  target variable         */
level=,                                    /*                  target measurement level*/
segment=
);
   %local mname ;%let mname= (aa_model_getVarMeta) ;

    %let vvnFlag = %sysfunc(NVALID(1));
    options validvarname=ANY;

    %let syscc=0;
    proc contents data=&data out=&output(rename=(Type=NTYPE)) noprint;
    run;

    %if ^&vvnFlag %then %do;
        options validvarname = V7;
    %end;

    %if ("&SYSCC" ^= "0") %then %do;
       %goto endgmt;
    %end;

    data &output;
       set &output;
       length UNAME $32 ROLE $32 LEVEL $10 TYPE $1 ORDER $10 VARNUM 8;
       UNAME=upcase(NAME);
       if NTYPE=1 then do;
          Type = 'N';
          Level = 'INTERVAL';
        end;
        else do;
           Type='C';
           Level='NOMINAL';
        end;
        ROLE = 'INPUT';
        %if "&target" ne "" %then %do;
            if UNAME = upcase("&target") then do;
               ROLE = 'TARGET';
               %if "&level" ne "" %then %do;
                   Level = "&level";
               %end;
            end;
        %end;
        %if "&segment" ne "" %then %do;
            if UNAME = upcase("&segment") then do;
               ROLE = 'SEGMENT';
               Level = 'NOMINAL';
            end;
        %end;

        keep NAME UNAME ROLE LEVEL ORDER TYPE LABEL FORMAT INFORMAT LENGTH VARNUM;
    run;

    proc sort data=&output;by UNAME;run;

    %if ("&itemStoreVar" ne "") and %sysfunc(exist(&itemStoreVar)) %then %do;
        proc sort data=&itemStoreVar;by UNAME;run;
        data &output;
           merge &output(in=_a) &itemStoreVar(keep=UNAME AARole AALevel rename=(AARole=ROLE AALevel=Level));
           by UNAME;
           if _a then output;
        run;
    %end;


    %if ^%sysfunc(exist(&output)) %then %do;
        %put %sysfunc(sysmsg());
        data _null_;
           putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp,  model_noinputmetadata_error, NOQUOTE))";
           call symput('aaModelErrorString', 'model_noinputmetadata_error');
        run;
    %end;

    %endgmt:

%mend aa_model_getVarMeta;

/*====================================================================================*/
/* NAME:  aa_model_getInputTable                                                         */
/* BY:    sasdzl                                                                      */
/* DATE:  January 10, 2012                                                            */
/*------------------------------------------------------------------------------------*/
/* Determine scoring inputs based on score code type                                  */
/*====================================================================================*/
%macro aa_model_getInputTable(
/* Required Information ---------------------------------------------------------------*/
data=,                                     /*                  input-training data     */
scorecodeFile=,                            /* file name containing the score code      */
scorecodeFormat=DATASTEP,                  /* format of score code                     */
output=,                                   /* name of generated output table           */
/* Optional Information                  ----------------------------------------------*/
varmeta=,                                  /* basic metadata about variables           */
FCMPData=                                  /* Name of the FCMP data set                */
);
   %local mname ;%let mname= (aa_model_GETINPUTTABLE) ;
   %let _data = &data;

   %if "&output" ne "" and %sysfunc(exist(&output)) %then %do;
       proc delete data=&output;
       run;
   %end;

   %let it_varmeta = &varmeta;
   %if ^%sysfunc(exist(&it_varmeta)) or ("&it_varmeta" eq "") %then %do;
       %let it_varmeta = work._itvarmeta;
       %aa_model_getVarMeta(data=&_data, output=&it_varmeta);

       %if ^%sysfunc(exist(&it_varmeta)) %then %do;
       %goto errgit;
       %end;
   %end;

   %if %upcase("&scorecodeFormat") ne "DATASTEP" %then %do;
       %goto endgit;
    %end ;

   /*---------------------------------------------------------------
   * Call FCMP for dependencies
   *---------------------------------------------------------------*/
   %let _fcmpData = &fcmpData;
   %if "&fcmpData" eq "" %then
       %let _fcmpData = WORK._alldeps;


   %let SYSCC= 0;

   %put;
   %put NOTE: &mname: Checking Code for variable dependencies.;
   PROC FCMP  NOEXECUTE IMPUTEDDEPS OUTDEPS=&_fcmpData;
      %INCLUDE "&scoreCodeFile";
   RUN;
   QUIT;

   %if ("&SYSERR" ^= "0") %then %do;
       data _null_;
          putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_codedependency_error, NOQUOTE))";
          call symput('aaModelErrorString', 'model_codedependency_error');
       run;
       %put %sysfunc(sysmsg());
       %goto endgit;
   %end;

   %let inputNum = 0;
   /* check for empty _alldeps */
   data _NULL_;
      numobs=0;
      dsid=OPEN("&_fcmpData","i");
      if dsid then do;
         numobs= attrn(dsid,"nlobs");
      end;
      else do;
         putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_dependencydataerror_error, NOQUOTE))";
         call symput('aaModelErrorString', 'model_dependencydataerror_error');
         msg= sysmsg();
         numobs = -1;
      end;
      if numobs <= 0 then do;
         putlog "NOTE: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_nodependecyvariables_note, NOQUOTE))";
      end;
      dsid = close(dsid);
      call symput('inputNum', strip(put(numobs, best.)));
   Run;

   %if ("&SYSCC" ^= "0") %then %do;
       %goto endgit;
   %end;

   %if %sysfunc(exist(&_fcmpData)) %then %do;
       data work._Rhs (keep=rhs RHS_NLiteral UNAME);
          set &_fcmpData;
          Length UNAME $32;
          LABEL rhs = ;
          LABEL RHS_NLiteral = ;
          if RHS_Dropped = 0 AND (1 NE index(rhs, '_OLD_')) then do;
             UNAME = UPCASE(rhs);
             output;
          end;
       run;
       %if ^%sysfunc(exist(WORK._Rhs)) %then %do;
           %put %sysfunc(sysmsg());
           data _null_;
              putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_createintermediatefile_error, NOQUOTE))";
              call symput('aaModelErrorString', 'model_createintermediatefile_error');
           run;
           %goto endgit;
       %end;

       /* use proc sort to remove dups and get right hand side vars */
       proc sort Data=WORK._Rhs nodupkey out=WORK._Rhs (rename=(rhs=NAME  RHS_NLiteral=NLITERAL));
          by UNAME;
       run;
       %if ^%sysfunc(exist(WORK._Rhs)) %then %do;
            %put %sysfunc(sysmsg());
            data _null_;
               putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_createintermediatefile_error, NOQUOTE))";
               call symput('aaModelErrorString', 'model_createintermediatefile_error');
            run;
            %goto endgit;
       %end;

        proc sort data=&it_varmeta nodupkey;
           by UNAME;
        run;
        data &output;
           merge work._rhs(in=_a) &it_varmeta(in=_b);
           by UNAME;
           drop UNAME;
           length ROLE $32 LEVEL $10;
           if _a and _b then output;
        run;

   %end;

   %endgit:

   %if ^%sysfunc(exist(&output)) %then %do;
       data &output;
          set &it_Varmeta;
       run;
   %end;

   %if "&it_varmeta" ne "&varmeta" %then %do;
       proc delete data=&it_varmeta;
       run;
   %end;

   %errgit:

   %if ("&fcmpData" eq "") and %sysfunc(exist(&_fcmpData)) %then %do;
       proc delete data=&_fcmpData;
       run;
   %end;
%mend aa_model_getInputTable;

/*====================================================================================*/
/* NAME:  aa_model_getOutputTable                                                     */
/* BY:    sasdzl                                                                      */
/* DATE:  January 10, 2012                                                            */
/*------------------------------------------------------------------------------------*/
/* Determine scoring outputs                                                          */
/*====================================================================================*/
%macro aa_model_getOutputTable(
/* Required Information ---------------------------------------------------------------*/
data=,                                     /*                  input-training data     */
scorecodeFile=,                            /* file name containing the score code      */
scorecodeFormat=DATASTEP,                  /* format of score code                     */
scoredata=,                                /* name of the scored table                 */
output=,                                   /* name of generated output table           */
/* Optional Information                  ----------------------------------------------*/
varmeta=,                                  /* basic metadata about variablesin data    */
FCMPData=                                  /* Name of the FCMP data set                */
);

   %let got_scoreoutput = _score_scoreout;
   %let it_varmeta = &varmeta;
   %let ot_varmeta = _score_outvarmeta;

   %if ("&scoreCodeFile" ne "") and %sysfunc(fileexist(%nrbquote(&scoreCodeFile))) and ("&scoreCodeFormat" eq "DATASTEP") %then %do;
       data &got_scoreoutput;
          set &data;
          %inc "&scoreCodeFile";
       run;
   %end;
   %else %do;
       %if ("&scoredata" eq "") or (^%sysfunc(exist(&scoredata)) and ^%sysfunc(exist(&scoredata, VIEW))) %then %do;
           %goto exitgot;
       %end;
       %let got_scoreoutput = &scoredata;
   %end;


   %if ^%sysfunc(exist(&it_varmeta)) or ("&it_varmeta" eq "") %then %do;
       %let it_varmeta = work._model_itvarmeta;
       %aa_model_getVarMeta(data=&_data, type=INPUT, output=&it_varmeta);

       %if ^%sysfunc(exist(&it_varmeta)) %then %do;
           %goto exitgot;
       %end;
   %end;

   %aa_model_getVarMeta(data=&got_scoreoutput, type=OUTPUT, output=&ot_varmeta);
    %if ^%sysfunc(exist(&ot_varmeta)) %then %do;
       %goto exitgot;
   %end;

   %if %sysfunc(exist(&fcmpData)) and ("&scoreCodeFormat" eq "DATASTEP") %then %do;
       data work._lhs (keep=LHS LHS_NLiteral UNAME);
          set &_fcmpData;
          Length UNAME $32;
          LABEL rhs = ;
          LABEL RHS_NLiteral = ;
          if LHS_Dropped = 0 AND (1 NE index(LHS, '_OLD_')) then do;
             UNAME = UPCASE(LHS);
             output;
          end;
       run;
       proc sort Data=WORK._lhs nodupkey out=WORK._lhs (rename=(lhs=NAME  LHS_NLiteral=NLITERAL));
          by UNAME;
       run;
   %end;

   proc sort data=&it_varmeta(keep=UNAME) nodupkey;
      by UNAME;
   run;
   proc sort data=&ot_varmeta nodupkey;
      by UNAME;
   run;
   data &output;
      merge &it_varmeta(in=_a) &ot_varmeta(in=_b)
      %if %sysfunc(exist(work._LHS)) %then %do;
          work._LHS(in=_C)
      %end;
      ;
      by UNAME;
      if substr(UNAME, 1, 2)='P_' then ROLE='PREDICT';
      else
      if UNAME in('EM_PREDICTION', 'EM_EVENTPROBABILITY', 'EM_PROBABILITY') then ROLE = 'PREDICT';
      else
      if UNAME ='EM_CLASSIFICATION' then ROLE = 'CLASSIFICATION';
      else
      if UNAME ='EM_DECISION' then ROLE = 'DECISION';
      else
      if substr(UNAME, 1, 2)='R_' then ROLE='RESIDUAL';
      else
      if substr(UNAME, 1, 2) in('I_', 'F_', 'U_') then ROLE='CLASSIFICATION';
      else
      if substr(UNAME, 1, 2) ='Q_' then ROLE='ASSESS';
      else
      if UNAME in ('_WARN_', 'EM_PROFIT', 'EM_LOSS', 'EM_VALUETARGET', 'EM_CLASSTARGET') then ROLE='ASSESS';
      else
      if UNAME in ('_NODE_', 'EM_SEGMENT') then ROLE='SEGMENT';
      %if %sysfunc(exist(work._LHS)) %then %do;
          if length and (_c or _b) and ^_a then output;

      %end;
      %else %do;
         if _b and ^_a then output;
      %end;
   run;

   %exitgot:

   %let deleteString =;

   %if "&scoredata" eq "" %then %do;
       %if %sysfunc(exist(&got_scoreoutput)) %then %do;
            %let deleteString = &deleteString &got_scoreoutput;
       %end;
   %end;
   %if "&varmeta" eq "" and %sysfunc(exist(&it_varmeta)) %then %do;
       %let deleteString = &deleteString &it_varmeta;
   %end;

   proc datasets lib=work nowarn nolist;
      delete _rhs;
      delete _lhs;
      delete &ot_varmeta;
      %if "&deleteString" ne "" %then %do;
           delete &deleteString;
      %end;
   run;
   quit;


%mend aa_model_getOutputTable;

/*====================================================================================*/
/* NAME:  aa_model_getTableMeta                                                       */
/* BY:    sasdzl                                                                      */
/* DATE:  January 10, 2012                                                            */
/*------------------------------------------------------------------------------------*/
/* Determine metadata of training table                                               */
/*====================================================================================*/
%macro aa_model_getTableMeta(
/* Required Information ---------------------------------------------------------------*/
data=,                                     /*                  input-training data     */
output=                                    /* name of generated output table           */
);

    %let vvnFlag = %sysfunc(NVALID(1));
    options validvarname=ANY;

     data &output;
        length memname $ 32 memtype $ 8 memlabel $ 256 typemem $ 8 engine $ 8
               crdate 8 modate 8 nobs 8 ncols 8 USEEXTERNALDATA $1 SAMPLINGRATE 8 SEGMENT $ 20 LIBNAME $8;
        data = symget('data');
        dsid = open(data);
        libname = scan(data, 2, '.');
        if libname eq '' then libname = 'WORK';
        else libname = scan(data, 1, '.');
        if dsid then do;
           nobs      = attrn(dsid, 'nlobs');
           ncols     = attrn(dsid, 'nvars');
           memname   = attrc(dsid, 'mem');
           memlabel  = strip(attrc(dsid, 'label'));
           memtype   = attrc(dsid, 'mtype');
           typemem   = attrc(dsid, 'type');
           engine    = attrc(dsid, 'engine');
           crdate    = attrn(dsid,'crdte');
           modate    = attrn(dsid, 'modte');
           dsid =  close(dsid);
           useexternalData='';samplingrate=.;segment='';
        end;
        drop data dsid;
     run;

     %if ^&vvnFlag %then %do;
        options validvarname = V7;
    %end;

%mend aa_model_getTableMeta;

/*====================================================================================*/
/* NAME:  aa_SPKmsg                                                                   */
/* BY:    sasdzl                                                                      */
/* DATE:  January 10, 2012                                                            */
/*------------------------------------------------------------------------------------*/
/* Generate msg statement when creating spk file                                      */
/*====================================================================================*/
%macro aa_spk_msg;
     put 'else do;';
     put '   msg = sysmsg();';
     put '   put msg;';
     put 'end;';
%mend aa_spk_msg;

/*====================================================================================*/
/* NAME:  aa_createSPK                                                                */
/* BY:    sasdzl                                                                      */
/* DATE:  January 10, 2012                                                            */
/*------------------------------------------------------------------------------------*/
/* Create SPK with Model Information                                                  */
/*====================================================================================*/
/*====================================================================================*/
/* NAME:  aa_fromDStoXML                                                              */
/* BY:    sasdzl                                                                      */
/* DATE:  January 18, 2012                                                            */
/*------------------------------------------------------------------------------------*/
/* Convert a SAS data set to an XML file                                              */
/*====================================================================================*/
%macro aa_fromDStoXML(tag=, VarTable=, XMLFile=);

     %if "&vartable" eq "" or ^%sysfunc(exist(&varTable)) %then %do;
          %goto exitxml ;
     %end;

     %if "&tag" eq "" %then %let tag = AAMODEL;

     filename _emrprt "&XMLFile" nobom;
     libname _emrprt xml xmldataform=attribute encoding="UTF-8";
         data _emrprt.&tag;
            set &VarTable;
         run;
     libname  _emrprt;
     filename _emrprt;

     %exitxml:

%mend aa_fromDStoXML;

/*====================================================================================*/
/* NAME:  aa_fromDStoXML                                                              */
/* BY:    sasdzl                                                                      */
/* DATE:  January 18, 2012                                                            */
/*------------------------------------------------------------------------------------*/
/* Convert a SAS data set to an XML file                                              */
/*====================================================================================*/
%macro aa_makeMiningResultXML(
modelname=,                              /*                    model name              */
modeldesc=,                              /*                    model description       */
modelKey=,                               /*                    unique model identifier */
targetVarTable=,                         /*             Metadata of target and segment */
tableMeta=,                              /*                    Table Metadata          */
scorecodeFormat=        ,                /*                    DATASTEP or other       */
miningAlgorithm=,                           /* Algorithm Description e.g. DecisionTree */
miningFunction=,                           /* Mining Function e.g. Classification     */
XMLFile=,
itemstore=                                 /* Indicates if model has an itemstore     */
);
   /* Setting type of score code */
   %let _scoreCodeFormat= SAS Program;
   %if "&scoreCodeFormat" eq "DATASTEP" %then
       %let _scoreCodeFormat= Data Step;

   filename _emrprt "&XMLFile"  encoding="UTF-8" NOBOM;

  %let numTarget  = 0;
  %let numSegment = 0;
  data _null_;
     retain numTarget numSegment 0;
     set &targetVarTable end=eof;
     if upcase(ROLE) = 'TARGET' then numTarget+1;
     else
     if upcase(ROLE)='SEGMENT' then numSergment+1;
     if eof then do;
        call symput('numTarget',  strip(put(numTarget, best.)));
        call symput('numSegment', strip(put(numSegment, best.)));
     end;
  run;
  data _null_;
     set &tableMeta;
     length string $200 itemstore $1 modelKey scoreCodeFormat miningAlgorithm miningFunction numTarget numSegment $36;
     file _emrprt;
     modelKey        = symget('modelKey');
     scoreCodeFormat = symget('_scoreCodeFormat');
     miningAlgorithm = symget('miningAlgorithm');
     miningFunction  = symget('miningFunction');
     numTarget       = symget('numTarget');
     numSegment      = symget('numSegment');
     itemstore       = symget('itemstore');

     userid          = symget('sysuserid');
     string = '<?xml version="1.0" encoding="UTF-8" ?><_ROOT_   UUID="'!!strip(modelKey)!!'" ScoreCodeType="'!!
             strip(scoreCodeFormat)!!'"';
     put string;
     string = 'MiningFunction="'!!strip(miningFunction)!!'" MiningAlgorithm="'!!strip(miningAlgorithm)!!'" Userid="'!!
     strip(userid)!!'" ItemStore="'!!strip(itemstore)!!'" Library="'!!strip(libname)!!'" Table="'!!strip(memname)!!'" Segment="" SamplingRate="." UseExternalData=""';
     put string;
     string = 'NTarget="'!!strip(numTarget)!!'" NSegment="'!!strip(numSegment)!!'">';
     put string;
  run;

  %if &numTarget %then %do;
     data _null_;
        set &targetVarTable end=eof;
        length string $200;
        file _emrprt mod;
        if _N_=1 then do;
           put '<TARGETS>';
        end;
        string ='<TARGET Variable="'!!strip(name)!!'" Level="'!!strip(level)!!'" Type="'!!strip(type)!!
                '" Format="'!!strip(format)!!'" Order="'!!strip(order)!!'"/>';
        put string;
        if eof then do;
           put '</TARGETS>';
        end;
      run;
  %end;
  %if &numSegment %then %do;
     data _null_;
        set &targetVarTable end=eof;
        length string $200;
        file _emrprt mod;
        if _N_=1 then do;
           put '<SEGMENTS>';
        end;
        string ='<SEGMENT Variable="'!!strip(name)!!'" Level="'!!strip(level)!!'" Type="'!!strip(type)!!
                '" Format="'!!strip(format)!!'" Order="'!!strip(order)!!'"/>';
        put string;
        if eof then do;
           put '</SEGMENTS>';
        end;
      run;
  %end;

  data _null_;
     file _emrprt mod;
     put '</_ROOT_>';
  run;
  filename _emrprt;
%mend aa_makeMiningResultXML;

/*====================================================================================*/
/* NAME:  aa_makeVarSpkXML                                                            */
/* BY:    sasdzl                                                                      */
/* DATE:  January 18, 2012                                                            */
/*------------------------------------------------------------------------------------*/
/* Use the XML for both input and Output type vars                                    */
/*====================================================================================*/
%macro aa_makeVarSpkXML(
/* Required Information --------------------------------------------------------------*/
VarMeta=,                                  /* metadata data set                       */
TypeVars=,                                 /* type of variable INPUT or OUTPUT        */
fileref=,                                  /* fileref associated with the output file */
);
   %lOCAL num_vars dsid;

   /* get the number of vars described by the metadata */
   %let num_vars = 0;
   %let dsid = %sysfunc(open(&varmeta));
   %if &dsid %then %do;
       %let num_vars = %sysfunc(attrn(&dsid, NLOBS));
       %let dsid = %sysfunc(close(&dsid));
   %end;
   %else %do;
      %put %SYSFUNC(sysmsg());
      %return;
   %end;

   data _null_;
      set &VarMeta end=eof;
      retain  varindex 0;
      length keepflag $5 strbuf $ 1024 vindexc $12;
      strbuf = ' ';
      file &fileref mod;

      vindexc = put(varnum, BEST12.);

      if _N_ = 1 then          /* befor starting first var          */
      do;
                               /* select start tag for in or out    */
         if "&TypeVars" = "INPUT" then
            strbuf =  ' <Input count="' || "&num_vars" || '">';
         else
            strbuf =  ' <Output count="' || "&num_vars" || '">';

          put @3 strbuf;
      end;

      strbuf =  '<Variable index="'||strip(vindexc)||'">';
      put @6 strbuf;

      /* if validvarname=any variable name */
      if (NLITERAL = 1) then do;
         strbuf = '<Vname>';
         put @9 strbuf;
         strbuf = '<![CDATA['||trim(name)||']]>';
         put @12 strbuf;
         put @9'</Vname>';
      end;
      else do;
         strbuf = '<Name> '||strip(name)||' </Name>';
         put @9 strbuf;
      end;

      CtypeVal = 'C';
      NtypeVal = 'N';
      if strip(type) = 'C' then
         strbuf = '<Type length="'|| strip(length)||'"> '||strip(CtypeVal)||' </Type>';
      else
         strbuf = '<Type> '||strip(NtypeVal)||' </Type>';

      put @9 strbuf;

      if LengthN(LABEL) then do;
         put @9 '<Description>';
         strbuf = '<![CDATA['||strip(LABEL)||']]>';
         put @12 strbuf;
         put @9 '</Description>';
      end;

      put @6 '</Variable>';

      varindex = varindex +1;

       /* put last */
      if eof then do;
         if "&TypeVars" = "INPUT" then
            strbuf =  "</Input>";
         else
            strbuf =  "</Output>";
         put @3 strbuf;
      end;
   Run;
%mend aa_makeVarSpkXML;

/*====================================================================================*/
/* NAME:  aa_makeSPKScoreXML                                                          */
/* BY:    sasdzl                                                                      */
/* DATE:  January 18, 2012                                                            */
/*------------------------------------------------------------------------------------*/
/*  generates an XML file to describe the metadata for the input and output variables */
/*  in the specified data step code and input data set.                               */
/*====================================================================================*/
%macro aa_makeSPKScoreXML(
/* Required Information ---------------------------------------------------------------*/
inputMeta =,                            /* data set with input  variable metadata      */
outputMeta=,                            /* data set with output variable metadata      */
XMLFile=                                /* output file containing the XML              */
);

   filename _emrprt "&XMLFile"  encoding="UTF-8" NOBOM;

   data _null_;
      LENGTH strbuf $ 80;
      strbuf = ' ';
      file _emrprt;
      put '<?xml version="1.0" encoding="utf-8"?>';
      put '<Score>';
      put @3 '<Producer>';
      put @6 "<Name>SAS: &sysuserid </Name>";
      strbuf = '<Version> 1.0 </Version>';
      put @6 strbuf;
                               /* get POSIX style locale            */
      syslocale = getpxlocale();
      strbuf = '<Locale> '|| strip(syslocale) ||' </Locale>';
      put @6 strbuf;
                               /* get SAS 12 byte encoding name     */
      sysencode = getpxencoding('sas12');
      strbuf = '<Encoding> '|| strip(sysencode) ||' </Encoding>';
      put @6 strbuf;

      put @3 '</Producer>';
   run;

   %aa_makeVarSpkXML(VarMeta=&inputMeta, typeVars=INPUT, fileref=_emrprt);
   %if ("&SYSCC" ^= "0") %then %do;
       data _null_;
          putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_inputvariablexml_error, NOQUOTE))";
          call symput('aaModelErrorString', 'model_inputvariablexml_error');
       run;
      %goto errexit;
   %end;

   %aa_makeVarSpkXML(VarMeta=&outputMeta, typeVars=OUTPUT, fileref=_emrprt);
   %if ("&SYSCC" ^= "0") %then %do;
       data _null_;
          putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_outputvariablexml_error, NOQUOTE))";
          call symput('aaModelErrorString', 'model_outputvariablexml_error');
       run;
      %goto errexit;
   %end;

   data _null_;
      file _emrprt mod;
      put '</Score>';
   run;

   filename _emrprt;

   %errexit:

%mend aa_makeSPKScoreXML;


%macro aa_insertSPKFile(
packageId=,
fileref=,
filename=,
filetype=,
mimetype=,
filedesc=,
name=
);

   string = "filename  &fileref"!!" '"!!"&filename"!!"'"!!';';
   put string;
   string = "call insert_file(&packageId,"!!'"fileref:'!!"&fileref"!!'", "'!!"&filetype"!!'", "'!!"&mimeType"!!'", '!!
                                "'"!!"&filedesc"!!"', 'NAME="!!"&name"!!"', rc);";
   put string;
   string='if rc eq 0 then put "File '!!"&filename"!!' inserted successfully.";';
   put string;
   put 'else do;';
   put '   msg = sysmsg();';
   put '   put msg;';
   put 'end;';
%mend aa_insertSPKFile;


%macro aa_createSPK(
/* Required Information ---------------------------------------------------------------*/
modelname=,                              /*                    model name              */
modeldesc=,                              /*                    model description       */
modelKey=,                               /*                    unique model identifier */
spkFolder=,                               /*            folder where to create spk file */
inputVarTable=,                          /*                Metadata of scoring inputs  */
outputVarTable=,                         /*             Metadata of scoring outputs    */
targetVarTable=,                         /*             Metadata of target and segment */
tableMeta=,                              /*                    Table Metadata          */
scorecodeFile=,                          /*        file name containing the score code */
scorecodeFormat=DATASTEP,                /*                    DATASTEP or other       */
/* Optional File                         ----------------------------------------------*/
trainFile=,                              /*  File name containing the training code    */
pmmlFile=,                               /*  file name containing the pmml score code  */
fitstats=,                               /* data set name with fit statistics          */
/* Optional Information  about the Model ----------------------------------------------*/
miningAlgorithm=,                           /* Algorithm Description e.g. DecisionTree */
miningFunction=,                            /* Mining Function e.g. Classification     */
debug=,
itemstore=                        /* Indicates if the model info came from an itemstore */

);

    %let workpath = %sysfunc(pathname(WORK));
    %if "&spkFolder" eq "" %then %do;
        %let spkFolder = &workpath;
    %end;
    %let spkFolder = %superq(spkFolder)&msdsep&modelkey;
    %let aa_spk_folderid= %superq(spkFolder);

    %let inputxml  = &workpath&msdsep.input.xml;
    %let outputxml = &workpath&msdsep.output.xml;
    %let targetxml = &workpath&msdsep.target.xml;
    %let tablexml  = &workpath&msdsep.table.xml;
    %let miningResultxml  = &workpath&msdsep.MiningResult.xml;
    %let scorexml   = &workpath&msdsep.score.xml;
    %let fitstatsxml   = &workpath&msdsep.fitstats.xml;

    %let _modelName = %superq(modelName);
    %let _modelDesc = %superq(modelDesc);
    %let _modelKey        = &modelKey;
    %let _targetVarTable  = &targetVarTable;
    %let _tableMeta       = &tableMeta;
    %let _scoreCodeFormat = &scoreCodeFormat;
    %let _miningFunction  = &miningFunction;
    %let _miningAlgorithm = &miningAlgorithm;
    %let _trainFile       = &trainFile;
    %let _fitstats        = &fitstats;
    %let _itemstore       = &itemstore;


    %aa_makeMiningResultXML(modelname=%superq(_modelName), modeldesc=%superq(_modelDesc), modelKey=%superq(_modelKey),
                    targetVarTable=&_targetVarTable,  tableMeta=&_tableMeta, scorecodeFormat=&_scoreCodeFormat,
                     miningFunction=&_miningFunction, miningAlgorithm=&_miningAlgorithm, XMLFile=&miningResultxml,
                     itemstore=&_itemstore);

    %aa_fromDStoXML(tag=INPUT,     VarTable=&inputVarTable,  XMLFile=&inputxml);
    %aa_fromDStoXML(tag=OUTPUT,    VarTable=&outputVarTable, XMLFile=&outputxml);
    %aa_fromDStoXML(tag=TARGET,    VarTable=&targetVarTable, XMLFile=&targetxml);
    %aa_fromDStoXML(tag=ROOTMETA,  VarTable=&tableMeta,      XMLFile=&tablexml);
    %aa_fromDStoXML(tag=STATISTIC, VarTable=&_fitstats,      XMLFile=&fitstatsxml);


    %aa_makeSPKScoreXML(inputMeta = &inputVarTable, outputMeta= &outputVarTable, XMLFile=&scoreXML);

    %let sourceEntry = work.aa_createSPK.code.source;
    filename _fspk1 catalog "&sourceEntry";
    data _null_;
       length string $500
       spkpath spkfolder $400
       modelname modeldesc $200 modelkey $36;
       modelname = symget('_modelname');
       modeldesc = symget('_modeldesc');
       modelkey  = symget('_modelkey');
       spkfolder = symget('_spkfolder');
       spkpath = dcreate(modelkey, spkFolder);
       file _fspk1;
       put 'data _null_;';
       put 'rc = 0;';
       put 'Pack_1 = 0;';
       string = 'call package_begin(Pack_1, "'!!strip(ModelName)!!'", '!!"'"!!' type="Model Result" version="12.1" Report="'!!
                strip(modelkey)!!'"'!!' description="'!!strip(modeldesc)!!'"'!!"', rc);";
       put string;
       string= 'if rc eq 0 then put "Package '!!strip(modelname)!!' began successfully.";';
       put string;
       %aa_spk_msg;

       put 'Pack_2 = 0;';
       string = 'call package_begin(Pack_2, "MININGRESULT", '!!"' NAME="!!'"Mining Result Metadata"'!!"', rc);";
       put string;
       put 'if rc eq 0 then put "Package MININGRESULT began successfully.";';
       %aa_spk_msg;

       string ='call insert_package(Pack_1, Pack_2, rc, "NESTED_NAME", "MININGRESULT");';
       put string;
       put 'if rc eq 0 then put "spk_PACKAGEINSERT_note";';
       %aa_spk_msg;

       %let miningXMLFlag = %sysfunc(fileexist(&miningResultxml));
       %if &miningXMLFlag %then %do;
           %aa_insertSPKFile(packageId=PACK_2, fileref=_fspk2, filename=%superq(miningResultxml), filetype=BINARY, mimetype=text/xml,
                             filedesc=MININGRESULT, name=MININGRESULT TYPE=BINARY);

       %end;

       %let inputXMLFlag = %sysfunc(fileexist(&inputxml));
       %if &inputXMLFlag %then %do;
           %aa_insertSPKFile(packageId=PACK_2, fileref=_fspk3, filename=%superq(inputxml), filetype=BINARY, mimetype=text/xml,
                             filedesc=INPUTTABLE, name=INPUTTABLE TYPE=BINARY);
       %end;

       %let outputXMLFlag = %sysfunc(fileexist(&outputxml));
       %if &outputXMLFlag %then %do;
           %aa_insertSPKFile(packageId=PACK_2, fileref=_fspk4, filename=%superq(outputxml), filetype=BINARY, mimetype=text/xml,
                             filedesc=OUTPUTTABLE, name=OUTPUTTABLE TYPE=BINARY);
       %end;

       %let targetXMLFlag = %sysfunc(fileexist(&targetxml));
       %if &targetXMLFlag %then %do;
           %aa_insertSPKFile(packageId=PACK_2, fileref=_fspk5, filename=%superq(targetxml), filetype=BINARY, mimetype=text/xml,
                             filedesc=TARGETTABLE, name=TARGETTABLE TYPE=BINARY);
       %end;

       %let tableXMLFlag = %sysfunc(fileexist(&tablexml));
       %if &tableXMLFlag %then %do;
           %aa_insertSPKFile(packageId=PACK_2, fileref=_fspk6, filename=%superq(tablexml), filetype=BINARY, mimetype=text/xml,
                             filedesc=TABLEMETA, name=TABLEMETA TYPE=BINARY);
       %end;

       %let fitstatsXMLFlag = %sysfunc(fileexist(&fitstatsxml));
       %if &fitstatsXMLFlag %then %do;
           %aa_insertSPKFile(packageId=PACK_2, fileref=_fspk11, filename=%superq(fitstatsxml), filetype=BINARY, mimetype=text/xml,
                             filedesc=FITSTAT, name=FITSTAT TYPE=BINARY);
       %end;

       %let scoreXMLFlag = %sysfunc(fileexist(&scorexml));
       %if &scoreXMLFlag %then %do;
           %aa_insertSPKFile(packageId=PACK_2, fileref=_fspk7, filename=%superq(scorexml), filetype=BINARY, mimetype=text/xml,
                             filedesc=SML, name=SCOREXML TYPE=BINARY);
       %end;

       %let trainFileFlag = %sysfunc(fileexist(&trainFile));
       %if &trainFileFlag %then %do;
           %aa_insertSPKFile(packageId=PACK_2, fileref=_fspk10, filename=%superq(trainfile), filetype=TEXT, mimetype=text/plain,
                             filedesc=TRAIN, name=TRAINCODE TYPE=TEXT);
       %end;

       put 'Pack_3 = 0;';
       string = 'call package_begin(Pack_3, "PATHSCORE", '!!"' NAME="!!'"Path score code"'!!"', rc);";
       put string;
       put 'if rc eq 0 then put "Package PATHSCORE began successfully.";';
       %aa_spk_msg;

       string ='call insert_package(Pack_1, Pack_3, rc, "NESTED_NAME", "PATHSCORE");';
       put string;
       put 'if rc eq 0 then put "spk_PACKAGEINSERT_note";';
       %aa_spk_msg;

       put 'Pack_4 = 0;';
       string = 'call package_begin(Pack_4, "SASSCORE", '!!"' NAME="!!'"SAS Score Code"'!!"', rc);";
       put string;
       put 'if rc eq 0 then put "Package SASSCORE began successfully.";';
       %aa_spk_msg;

       string ='call insert_package(Pack_3, Pack_4, rc, "NESTED_NAME", "SASSCORE");';
       put string;
       put 'if rc eq 0 then put "spk_PACKAGEINSERT_note";';
       %aa_spk_msg;

       %let scorecodeFlag = %sysfunc(fileexist(&scorecodeFile));
       %if &scoreCodeFlag %then %do;
           %aa_insertSPKFile(packageId=PACK_4, fileref=_fspk8, filename=&scorecodeFile, filetype=TEXT, mimetype=text/plain,
                             filedesc=SASCODE, name=SASCODE TYPE=TEXT);
       %end;

       put 'Pack_5 = 0;';
       string = 'call package_begin(Pack_5, "PMMLSCORE", '!!"' NAME="!!'"PMML Score Code"'!!"', rc);";
       put string;
       put 'if rc eq 0 then put "Package PMMLSCORE began successfully.";';
       %aa_spk_msg;

       string ='call insert_package(Pack_3, Pack_5, rc, "NESTED_NAME", "PMMLSCORE");';
       put string;
       put 'if rc eq 0 then put "spk_PACKAGEINSERT_note";';
       %aa_spk_msg;

       %let pmmlFlag = %sysfunc(fileexist(&pmmlFile));
       %if &pmmlFlag %then %do;
           %aa_insertSPKFile(packageId=PACK_5, fileref=_fspk9, filename=&pmmlFile, filetype=BINARY, mimetype=text/xml,
                             filedesc=PMMLCODE, name=PMML TYPE=BINARY);

       %end;
       put 'pubType = "TO_ARCHIVE";';
       put "properties='archive_path, archive_name';";
       string = 'path = "'!!strip(spkpath)!!'";';
       put string;
       put 'name = "miningResult";';
       put "call package_publish(Pack_1, pubType, rc, properties, path, name);";
       put 'if rc eq 0 then put "Publish package Pack_1 successfully.";';
       %aa_spk_msg;

       put "call package_end(Pack_1, rc);";
       put 'if rc eq 0 then put "Package Pack_1 ended successfully.";';

       %aa_spk_msg;
       put 'run;';
       put 'data _null_;';
       %if &miningXMLFlag %then %do;
           put 'filename _fspk2;';
       %end;
       %if &inputXMLFlag %then %do;
           put 'filename _fspk3;';
       %end;
       %if &outputXMLFlag %then %do;
           put 'filename _fspk4;';
       %end;
       %if &targetXMLFlag %then %do;
           put 'filename _fspk5;';
       %end;
       %if &tableXMLFlag %then %do;
           put 'filename _fspk6;';
       %end;
       %if &scoreXMLFlag %then %do;
           put 'filename _fspk7;';
       %end;
       %if &scoreCodeFlag %then %do;
           put 'filename _fspk8;';
       %end;
       %if &pmmlFlag %then %do;
           put 'filename _fspk9;';
       %end;
       %if &trainFileFlag %then %do;
           put 'filename _fspk10;';
       %end;
       %if &fitstatsXMLFlag %then %do;
           put 'filename _fspk11;';
       %end;

       put 'run;';
    run;

    %inc _fspk1;
    filename _fspk1;


%mend aa_createSPK;

/*====================================================================================*/
/* NAME:  aa_registerToMetadataServer                                                 */
/* BY:    sasdzl                                                                      */
/* DATE:  January 10, 2012                                                            */
/*------------------------------------------------------------------------------------*/
/* Register model to Metadata Server                                                  */
/*====================================================================================*/
%macro aa_registerToMetadataServer(
/* Required Information ---------------------------------------------------------------*/
modelname=,                                  /*                model name              */
modeldesc=,                                  /*                model description       */
modelKey=,                                   /*                unique model identifier */
mrPath=,                                     /*                model repository folder */
inputVarTable=,                              /*            Metadata of scoring inputs  */
outputVarTable=,                             /*            Metadata of scoring outputs */
targetVarTable=,                             /*         Metadata of target and segment */
tableMeta=,                                  /*                Table Metadata          */
scorecodeFile=,                              /*    file name containing the score code */
scorecodeFormat=DATASTEP,                   /*                 DATASTEP or other       */
/* Optional Information  about the Model ----------------------------------------------*/
miningAlgorithm=,                           /* Algorithm Description e.g. DecisionTree */
miningFunction=,                            /* Mining Function e.g. Classification     */
itemstore=
);

   %local mname ;%let mname= (aa_registerToMetadataServer) ;

   %let requestFile  = %sysfunc(pathname(WORK))&msdsep.request.xml;
   %let responseFile = %sysfunc(pathname(WORK))&msdsep.response.xml;

   filename response  "&responsefile";
   filename request   "&requestfile";

   %global FoundationRepositoryFullId;
   %global FoundationRepositoryId;

   /* Get Repository ID */
   data _null_;
      file request LRECL=5000;
      put '<GetRepositories><Repositories/><Flags>0</Flags><Options/></GetRepositories>';
      run;
   proc metadata in=request out=response;
   run;
   data _null_;
      infile response length=len lrecl=32767;
      length card $ 32767 repxml $ 255 reposid $ 17;
      input card $varying32767. len;
      do i=1 to 20;
         startrep=index(card,'<Repository');
         endrep=index(card,'/>');
         replen=(endrep-startrep+2);
         repxml=substr(card,startrep,replen);
         ndx=index(repxml,"Foundation");
         if ndx>0 then do;
            ndx=index(repxml,' Id=');
            repxml=substr(repxml,ndx);
            reposid=scan(repxml,2,'"');
            call symput('FoundationRepositoryFullId',reposid);
            call symput('FoundationRepositoryId',scan(reposid,2,'.'));
            stop;
         end;
         card=substr(card,endrep+1);
      end;
   run;

   /* Get folder ID */
   data _null_;
     length path type uri_t folder uri_f $256;
     * find an object;
     repos  = '';path   = '';deftype= '';
     type   = '';uri_t  = '';uri_f  = '';
     if "&mrpath" ne "" then do;
        folder = symget('mrpath');
        if substr(reverse(trim(folder)), 1, 1) ne '/' then
            folder = trim(folder)!!'/';
        rc_folder = metadata_pathobj(repos, folder, deftype, type, uri_f);

        if rc_folder le 0 then do;/* if all levels not found */
           uri_f = '';
        end ;
     end;
     else do;
        folder='/Shared Data/';
        rc_folder=metadata_pathobj(repos,folder,deftype,type,uri_f);
     end;
     call symput('FolderId', strip(uri_f));
  run;

  %if "&FolderId" eq "" %then %do;
      data _null_;
           putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_repositorynotfound_error, NOQUOTE, &FolderId))";
           call symput('aaModelErrorString', 'model_repositorynotfound_error');
      run;
      %goto exitrtm ;
  %end;

   /* Setting type of score code */
   %let ms_scoreCodeFormat= SAS Program;
   %if "&scoreCodeFormat" eq "DATASTEP" %then
       %let ms_scoreCodeFormat= Data Step;

   data _null_;
     file request LRECL=5000;
     length MODEL_NAME MODEL_DESC $200;
     modelkey = strip("&modelKey");
     MODEL_NAME=tranwrd("&modelname", '&', '&amp;');
     MODEL_NAME=tranwrd(MODEL_NAME, '"', '&quot;');
     MODEL_NAME=tranwrd(MODEL_NAME, '<', '&lt;');
     MODEL_NAME=tranwrd(MODEL_NAME, '>', '&gt;');
     MODEL_NAME=tranwrd(MODEL_NAME, "'", ' &#039;');

     MODEL_DESC=tranwrd("&modeldesc", '&', '&amp;');
     MODEL_DESC=tranwrd(MODEL_DESC, '"', '&quot;');
     MODEL_DESC=tranwrd(MODEL_DESC, '<', '&lt;');
     MODEL_DESC=tranwrd(MODEL_DESC, '>', '&gt;');
     MODEL_DESC=tranwrd(MODEL_DESC, "'", ' &#039;');

     length xml $5000;
       xml='<AddMetadata>';
      put xml;
      xml='<Metadata>';
      put xml;
      xml='<MiningResult Name="'!!strip(model_name)!!'" Desc="'!!strip(model_desc)!!'" usageVersion="1000000" PublicType="MiningResults"'!!
          ' ModelKey="'!!"&modelKey"!!'" MiningFunction="'!!"&miningFunction"!!'" MiningAlgorithm="'!!"&miningAlgorithm"!!'" ItemStore="'!!"&itemStore"!!'" SourceCodeType="'!!"&ms_scoreCodeFormat"!!'" Subject="No subject" Rating="0">';
      put xml;
      xml='<Extensions>';
      put xml;
      xml='<Extension Name="Developer" Value="'!!"&sysuserid"!!'"/>';
      put xml;
      timeval  = left(putn(time(), 'time.'));
      date     = left(putn(date(), 'YYMMDDD10.'));
      time     = trim(date)||" "||trim(timeval);
      xml='<Extension Name="Time" Value="'!!strip(time)!!'" />';
      put xml;
      xml='<Extension Name="Id" Value="id" />';
      put xml;
      xml='</Extensions>';
      put xml;
     run;

    /* Score Code */
    %if %sysfunc(fileexist(%nrbquote(&scoreCodeFile)))%then %do;
        filename _scrfile "%nrbquote(&scoreCodeFile)";
        data _null;
           length text $5000;
            file request mod LRECL=5000;
           infile _scrfile;
           put '<SourceCode Name="pathScoreCode.txt" >';
           put '<TextStore Name="Score Code" TextRole="Score" StoredText="';
           fileid=fopen('_scrfile', 'i', 1000);
           if (fileid>0) then do;
                do while(fread(fileid)=0);
                    rc=fget(fileid, text, frlen(fileid));
                    text=tranwrd(text, '&', '&amp;');
                    text=tranwrd(text, '"', '&quot;');
                    text=tranwrd(text, '<', '&lt;');
                    text=tranwrd(text, '>', '&gt;');
                    text = strip(text)!!' &#10;';
                    put text;
                end;
                rc=fclose(fileid);
            end;
            put '"/>';
            put '</SourceCode>';
        run;
        filename _scrfile;
    %end;

    /* Training Table */
    %if %sysfunc(exist(&tableMeta)) %then %do;
        data _null_;
           length string $200;

           set &tableMeta;

           file request mod LRECL=5000;
           put '<TrainingTable>';
           string ='<PhysicalTable Name="'!!strip(memname)!!'" SASTableName="'!!strip(memname)!!
                 '" MemberType="'!!strip(memtype)!!'" Desc="'!!strip(put(ncols, best.))!!'" NumRows="'!!
           strip(put(nobs, best.))!!'">';
           put string;
           put '<AnalyticTables>';
           string='<AnalyticTable Name="'!!strip(memname)!!'" Desc="'!!strip(libname)!!'" UseExternalData="'!!strip(UseExternalData)!!
                     '" SamplingRate="'!!strip(put(samplingRate, best.))!!'" Segment="'!!strip(segment)!!'" AnalyticTableRole="TRAIN"/>';
           put string;
           put '</AnalyticTables>';
           put '</PhysicalTable>';
          put '</TrainingTable>';
       run;
    %end;

    /* Target Table */
    %if %sysfunc(exist(&targetVarTable)) %then %do;
         data _null_;
            length string $5000 tempname templabel $200;
            set &targetVarTable;
            file request mod LRECL=5000;
            if _N_=1 then do;
                put '<TargetTable>';
                put '<RelationalTable Name="TargetTable">';
                put '<Columns>';

               tempname=tranwrd(NAME, '&', '&amp;');
               tempname=tranwrd(tempname, '"', '&quot;');
               tempname=tranwrd(tempname, '<', '&lt;');
               tempname=tranwrd(tempname, '>', '&gt;');
               tempname=tranwrd(tempname, "'", '&#39;');

               templabel=tranwrd(label, '&', '&amp;');
               templabel=tranwrd(templabel, '"', '&quot;');
               templabel=tranwrd(templabel, '<', '&lt;');
               templabel=tranwrd(templabel, '>', '&gt;');
               templabel=tranwrd(templabel, "'", ' &#039;');



               string='<Column Name="'!!trim(tempname)!!'" SASColumnName="'!!trim(tempname)!!'" SASColumnType="'!!strip(TYPE)!!
                      '" Desc="'!!strip(TEMPLABEL)!!'" SASFormat="'!!strip(FORMAT)!!'" SASInformat="'!!strip(INFORMAT)!!'"'!!
                      ' ColumnLength="'!!strip(put(LENGTH, best.))!!'" SASColumnLength="'!!strip(put(LENGTH, best.))!!'">';
               put string;
               put '<AnalyticColumns>';
               string = '<AnalyticColumn Level="'!!strip(LEVEL)!!'" Order="'!!strip(ORDER)!!'" NAME="'!!trim(TEMPNAME)!!'"/>';
               put string;
               put '</AnalyticColumns>';
               put '</Column>';
               put '</Columns>';
               put '</RelationalTable>';
               put '</TargetTable>';
            end;
         run;
     %end;

     /* Scoring Input Table */
     data _null_;
        file request mod LRECL=5000;
        put '<TransformationSources>';
        put '<RelationalTable Name="InputTable">';
        put '<Columns>';
     run;


     %if %sysfunc(exist(&inputvarTable)) %then %do;
         data _null_;
            set &inputvarTable;
            length string $2000 tempname templabel $200;
            file request mod LRECL=5000;

            tempname=tranwrd(NAME, '&', '&amp;');
            tempname=tranwrd(tempname, '"', '&quot;');
            tempname=tranwrd(tempname, '<', '&lt;');
            tempname=tranwrd(tempname, '>', '&gt;');
            tempname=tranwrd(tempname, "'", '&#39;');

            templabel=tranwrd(label, '&', '&amp;');
            templabel=tranwrd(templabel, '"', '&quot;');
            templabel=tranwrd(templabel, '<', '&lt;');
            templabel=tranwrd(templabel, '>', '&gt;');
            templabel=tranwrd(templabel, "'", ' &#039;');


            string='<Column Name="'!!trim(TEMPNAME)!!'" SASColumnName="'!!trim(TEMPNAME)!!'" SASColumnType="'!!strip(TYPE)!!
                      '" Desc="'!!trim(TEMPLABEL)!!'" SASFormat="'!!strip(FORMAT)!!'" SASInformat="'!!strip(INFORMAT)!!
                      '" ColumnLength="'!!strip(put(LENGTH, best.))!!'" SASColumnLength="'!!strip(put(LENGTH, best.))!!'">';
            put string;
            put '<AnalyticColumns>';
            string = '<AnalyticColumn Level="'!!strip(LEVEL)!!'" Order="'!!strip(ORDER)!!'" NAME="'!!trim(TEMPNAME)!!'"/>';
            put string;
            put '</AnalyticColumns>';
            put '</Column>';
         run;
     %end;

     data _null_;
        file request mod LRECL=5000;
        put '</Columns>';
        put '</RelationalTable>';
        put '</TransformationSources>';
     run;

     /* Output Table */
     data _null_;
        file request mod LRECL=5000;
        put '<TransformationTargets>';
        put '<RelationalTable Name="OutputTable">';
        put '<Columns>';
     run;

     %if %sysfunc(exist(&outputvarTable)) %then %do;
         data _null_;
            set &OutputVarTable;
            length string $2000 templabel tempname $200;
            file request mod LRECL=5000;

            tempname=tranwrd(NAME, '&', '&amp;');
            tempname=tranwrd(tempname, '"', '&quot;');
            tempname=tranwrd(tempname, '<', '&lt;');
            tempname=tranwrd(tempname, '>', '&gt;');
            tempname=tranwrd(tempname, "'", '&#39;');

            templabel=tranwrd(label, '&', '&amp;');
            templabel=tranwrd(templabel, '"', '&quot;');
            templabel=tranwrd(templabel, '<', '&lt;');
            templabel=tranwrd(templabel, '>', '&gt;');
            templabel=tranwrd(templabel, "'", ' &#039;');

            string='<Column Name="'!!trim(TEMPNAME)!!'" SASColumnName="'!!trim(TEMPNAME)!!'" SASColumnType="'!!strip(TYPE)!!
                      '" Desc="'!!trim(TEMPLABEL)!!'" SASFormat="'!!strip(FORMAT)!!'" SASInformat="'!!strip(INFORMAT)!!'" ColumnLength="'!!strip(put(LENGTH, best.))!!'" SASColumnLength="'!!strip(put(LENGTH, best.))!!'">';
            put string;
            put '<AnalyticColumns>';
            string = '<AnalyticColumn  AnalyticColumnRole="'!!strip(ROLE)!!'" Level="'!!strip(LEVEL)!!'" Order="'!!strip(ORDER)!!'" NAME="'!!trim(TEMPNAME)!!
                        '"/>';
            put string;
            put '</AnalyticColumns>';
            put '</Column>';
         run;
     %end;

     data _null_;
        file request mod LRECL=5000;
        put '</Columns>';
        put '</RelationalTable>';
        put '</TransformationTargets>';
     run;

     data _null;
         file request mod LRECL=5000;
         length xml $20000;
       xml='<Trees Function="Append">';
      put xml;
      xml= '<Tree ObjRef="'!!strip("&folderid")!!'"'!!'/>';
      put xml;
      xml= ' </Trees>';
      put xml;
      xml='</MiningResult>'!!'</Metadata>'!!"<Reposid>&FoundationRepositoryFullId"!!"</Reposid>"!!
         '<NS>SAS</NS>'!!'<Flags>268435456</Flags>';
      put xml;
      xml='<Options/>';
      put xml;
      xml='</AddMetadata>';
      put xml;
      run;

      proc METADATA
         in=request
         out=response;
      run;

     %let  aa_modelkey= &modelkey;

     %if %sysfunc(fexist(response)) %then %do;
        data _null_;
           length text $2000 miningresultid $32;
           infile response;
           input text $1-2000;
           text = scan(text, 3, '>');
           miningresultid = substr(text, index(text, 'Id="')+3);
           miningresultid = scan(miningresultid, 1, '"');
           call symput('aa_miningresultid', miningresultid);
        run;
     %end;

     filename request;
     filename response;

   %exitrtm:
%mend aa_registerToMetadataServer;


%macro aa_Model_register(
/* Required Information if not itemstore is specified ---------------------------------*/
data=,                                          /*             input-training data     */
target=,                                        /*             name of target variable */
level=,                                         /*             Measurement level       */
scorecodeFile=,                                 /* file name containing the score code */
scorecodeFormat=DATASTEP,                       /*             DATASTEP or other       */
/* Optional action                       ----------------------------------------------*/
register=Y,                                   /* register model to metadata server     */
mrPath=,                                      /* model repository folder               */
spk=N,                                        /* create an spk file                    */
spkFolder=,                                   /* folder where spk file will be created */
/* Optional Information  about the Model ----------------------------------------------*/
modelname=,                                 /* model name                              */
modeldesc=,                                 /* model description                       */
pmmlFile=,                                  /* file containing score code in pmml form */
lookup=,                                    /* algorithm for looking up class levels in*/
                                            /* the score code generate by PROC PLM     */
                                            /* AUTO (default), BINARY, SELECT, LINEAR  */
itemStore=,                                 /* item store generated by PLM procedure   */
score=,                                     /* Scored input-training data              */
inputvarsTable=,                            /* Metadata for the score inputs           */
outputvarsTable=,                           /* Metadata for the score outputs          */
trainFile=,                                 /* Training Program                        */
miningAlgorithm=,                           /* Algorithm Description e.g. DecisionTree */
miningFunction=,                            /* Mining Function e.g. Classification     */
/* Optional Information for other mining function than modeling e.g. Segmentation      */
segment=,                                                  /* name of Segment variable */
/* Debug Information when set Y prevent deletion of generated data sets                */
debug=N
);
   %global aa_spk_folderid aa_modelkey aa_miningresultid;
   %let aa_spk_folderid=;
   %let aa_modelkey=;
   %let aa_miningresultid=;

   %let aaModelErrorString=;

   %let syscc=0;
   %local mname ;%let mname= (aa_model_register) ;

   %if %upcase("&spk") eq "N" and %upcase("&register") eq "N" %then %do;
       data _null_;
          putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_noaction_error, NOQUOTE))";
          call symput('aaModelErrorString', 'model_noaction_error');
       run;
       %goto exit ;
   %end;

   /* Retrieve the correct file separator */
   %global msdep;
   %let msdsep=;
   %if %substr(&sysscp, 1, 3)= WIN %then
       %let msdsep=\;
   %else
       %if %substr(&sysscp, 1, 3)= DNT %then
       %let msdsep=\;
   %else
       %let msdsep=/;

   %let _data = &data;
   %let _scoreCodeFile = %superq(scoreCodeFile);

   %let _fitstats      = work._aafitstats;
   %let _itemStoreVar  = work._aaitemStoreVars;
   %let _itemStoreInfo = work._aaitemStoreInfo;
   %let _itemStorePgm  = %sysfunc(pathname(WORK))&msdsep.trainpgm.sas;
   %let _debug = &debug;


    %if "&spk" eq "Y" %then %do;
        %if "&spkFolder" eq "" %then %do ;
            data _null_;
               putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_nospkfoldername_error, NOQUOTE))";
               call symput('aaModelErrorString', 'model_nospkfoldername_error');
            run;
            %goto exit ;
        %end ;
    %end;

    %let _itemStoreVar=;
    %let _itemStoreFlag=N;


     /* If itemstore is not specified then check for required arguments */
    %if "&itemStore" eq "" %then %do;
        %if "&data" eq "" or (^%sysfunc(exist(&data)) and ^%sysfunc(exist(&data, VIEW))) %then %do;
            data _null_;
               putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_nodata_error, NOQUOTE, &data))";
               call symput('aaModelErrorString', 'model_nodata_error');
            run;
            %goto exit ;
        %end ;
        %if "%superq(target)" eq "" %then %do ;
            data _null_;
               putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_notarget_error, NOQUOTE))";
               call symput('aaModelErrorString', 'model_notarget_error');
            run;
            %goto exit ;
        %end ;
        %else %do;
            data _null_;
                     target = trim(symget('target'));
                     len = length(target);
                     if len>3 then do;

                        x = substr(target, 1, 1);
                  y= substr(reverse(trim(target)), 1, 2);
                  if (substr(target, 1, 1)="'" and substr(reverse(trim(target)), 1, 2) = "n'") or
                     (substr(target, 1, 1)='"' and substr(reverse(trim(target)), 1, 2) = 'n"') then do;
                               target = substr(target,2, len-3);
                     call symput('target', trim(target));
                  end;
                     end;
                  run;
        %end;

        %if "&level" eq "" %then %do ;
            data _null_;
               putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_notargetlevel_error, NOQUOTE))";
               call symput('aaModelErrorString', 'model_notargetlevel_error');
            run;
             %goto exit ;
        %end ;
        %if "&scoreCodeFile" eq "" %then %do ;
            data _null_;
               putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_noscorecodefile_error, NOQUOTE))";
               call symput('aaModelErrorString', 'model_noscorecodefile_error');
            run;
             %goto exit ;
        %end ;
    %end;
    %else %do;
        %let itemStore_lib=work;
        /* Check if item store exists */
        %let itemStore_name = %scan(&itemStore, 2, .);
        %if "&itemStore_name" eq "" %then %do;
            %let itemStore_lib = work;
            %let itemStore_name = %scan(&itemStore, 1, .);
        %end;
        %else %do;
            %let itemStore_lib  = %scan(&itemStore, 1, .);
            %let itemStore_name = %scan(&itemStore, 2, .);
        %end;
        %let itemStoreFile = %sysfunc(pathname(&itemStore_lib))&msdsep&itemStore_name..sas7bitm;
        %if ^%sysfunc(fileexist(&itemStoreFile)) %then %do;
            data _null_;
               putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_itemstorenotfound_error, NOQUOTE, &itemstoreFile))";
               call symput('aaModelErrorString', 'model_itemstorenotfound_error');
            run;
            %goto exit ;
        %end;

        %let _itemStore = &itemStore;
        %let _itemstoreFlag = Y;
        %let _itemStoreLookup = &lookup;

        %let _itemStoreVar = work._aaitemStoreVars;
        %let _itemStoreInfo = work._aaitemStoreInfo;
        %let trainFile       = %sysfunc(pathname(WORK))&msdsep.aa_traincode.sas;
        %let _scorecodeFile  = %sysfunc(pathname(WORK))&msdsep.aa_scorecode.sas;

        %let syscc= 0;
        %aa_itemstoreExtract(itemstore=&_itemStore, varDs=&_itemStoreVar, infoDs=&_itemStoreInfo, fitstats=&_fitstats,
                             pgmFile= %nrbquote(&trainFile), lookup=&_itemStoreLookup, scorecodeFile= %nrbquote(&_scorecodeFile), debug=&_debug);
        %if %eval(&syscc>4) %then %do;
            %goto exit ;
        %end;

        %if %sysfunc(exist(&_itemStoreInfo)) %then %do;
            data _null_;
               set &_itemStoreInfo;
               select(Name);
                   when('DATA')            call symput('_data', strip(value));
                   when('TARGET')          call symput('target', strip(value));
                   when('MININGFUNCTION')  call symput('miningfunction', strip(value));
                   when('MININGALGORITHM') call symput('miningalgorithm', strip(value));
                   otherwise;
               end;
            run;

        %end;
    %end;

    %if ^%sysfunc(fileexist(%superq(_scorecodefile))) %then %do;
        data _null_;
           putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_scorecodefilenotfound_error, NOQUOTE, &scorecodefile))";
           call symput('aaModelErrorString', 'model_scorecodefilenotfound_error');
        run;
        %goto exit ;
    %end;


   %let _modelname = %superq(modelname);
   %if "&_modelname" = "" %then %do;
       data _null_;
          putlog "NOTE: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_nomodelname_note, NOQUOTE, aa_model_&sysuserid))";
       run;
       %let _modelname = aa_model_&sysuserid;
   %end;

   %let _modeldesc = %superq(modeldesc);
   %if "&_modeldesc" = "" %then %do;
       %let _modeldesc = aa_model &sysuserid;
   %end;

   /* Setting MiningFunction if needed */
   %let _miningFunction = &miningFunction;
   %if "&_miningFunction" eq "" %then %do;
       %if %upcase("&level") eq "INTERVAL" %then
           %let _miningFunction = Prediction;
       %else
           %let _miningFunction = Classification;
   %end;
   %let _miningAlgorithm = &miningAlgorithm;

   %let _scoreCodeFormat= %upcase(&scoreCodeFormat);

   /* Unique Model Attributes */
   %let _modelKey = %sysfunc(UUIDGEN(1, 0));

   /* Training Table */
   %let _tablemeta = work._model_tableMeta;
   %aa_model_getTableMeta(data=&_data, output=&_tableMeta);

   /* Retrieving Variables Information from the Training Data */
   %let metadata = work._model_varsmeta;
   %let _target = %superq(target);
   %let _segment = %superq(segment);
   %let _level   = %upcase(&level);
   %aa_model_getVarMeta(data=&_data, output=&metadata, itemstoreVar= &_itemStoreVar, target=&_target, level=&_level, segment=&_segment);

  /* Create Target=Segment variable metadata table */
   %let _targetmeta = work._model_targetMeta;
   %let _numTargetSegment = 0;
   data &_targetMeta;
      set &metadata end=eof;
      where ROLE in('SEGMENT', 'TARGET');
      if eof then
         call symput('_numTargetSegment', strip(put(_n_, best.)));
    run;
   %if ^&_numTargetSegment %then %do;
       data _null_;
            if "&_Target" ne "" then do;
               putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_targetvarnotfound_error, NOQUOTE, &_target))";
               call symput('aaModelErrorString', 'model_targetvarnotfound_error');
            end;
            else
            if "&_segment" ne "" then do;
               putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_segmentvarnotfound_error, NOQUOTE, &_segment))";
               call symput('aaModelErrorString', 'model_segmentvarnotfound_error');
            end;
       run;
       %goto exit ;
   %end;

   /* Check if the all target and segment variables are in the data */
   %let _missVar = 0;
   %let _modelErrorVars=;
   data _null_;
      length missVar $200;
      retain dsid 0 missvar "";
      set &_targetMeta end=eof;
      if _n_=1 then do;
         dsid = open("&_data");
      end;
      if dsid then do;
           if ^varnum(dsid, NAME) then do;
                missVar = trim(missvar)!!' '!!NAME;
           end;
      end;
      if eof then do;
         if missVar ne '' then do;
            call symput('_modelErrorVars', trim(missVar));
            putlog "ERROR: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_varsnotfound_error, NOQUOTE, &_modelErrorVars))";
            call symput('aaModelErrorString', 'model_varsnotfound_error');
            call symput('_missVar', '1');
         end;
         if dsid then dsid = close(dsid);
      end;
   run;
   %if &_missVar %then %do;
        %goto exit ;
   %end;

   /* Create Scoring Input Table */
   %let _fcmpData       = work._model_fcmpDep;
   %let _inputVarTable = work._model_inputVarsTable;
   %if "&inputVarsTable" ne "" and %sysfunc(exist(&inputvarsTable)) %then %do;
       data &_inputVarTable;
          set &inputVarsTable;
       run;
   %end;
   %else %do;
       %aa_model_getInputTable(data=&_data, scoreCodeFile=&_scoreCodeFile, scorecodeFormat=&_scoreCodeFormat, varmeta=&metadata,
                              output=&_inputVarTable, fcmpData=&_fcmpData);
   %end;

   /* Create Scoring Output Table */
   %let _outputVarTable = work._model_outputVarsTable;
   %if "&outputVarsTable" ne "" and %sysfunc(exist(&outputvarsTable)) %then %do;
       data &_outputVarTable;
          set &outputVarsTable;
       run;
   %end;
   %else %do;
       %aa_model_getOutputTable(data=&_data, scoreCodeFile=&_scoreCodeFile, scorecodeFormat=&_scoreCodeFormat,
                      scoredata=&score, varmeta=&metadata, output=&_outputVarTable, fcmpData=&_fcmpData);
   %end;

   %if "&register" eq "Y" %then %do;
         %let _mrpath         = %superq(mrPath);
         %aa_registerToMetadataServer(modelname=%superq(_modelName), modeldesc=%superq(_modeldesc), modelkey=&_modelKey, mrPath=&_mrPath,
                inputVarTable= &_inputVarTable, outputVarTable=&_outputVarTable, tableMeta=&_tableMeta, targetVarTable=&_targetMeta, scorecodeFile=%superq(_scoreCodeFile),
                scorecodeFormat=&_scoreCodeFormat, miningFunction=&_miningFunction, miningAlgorithm=&_miningAlgorithm, itemStore=&_itemStoreFlag);

    %end;
    %if "&spk" eq "Y" %then %do;

        %let _pmmlFile=;
        %if "&pmmlFile" ne "" %then %do;
            %let _pmmlFile = %superq(pmmlFile);
            %if ^%sysfunc(fileexist(%superq(_pmmlfile))) %then %do;
                data _null_;
                           putlog "WARNING: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_pmmlfilenotfound_error, NOQUOTE, &_pmmlfile))";
                run;
            %end;
        %end;

        %let _trainFile = ;
        %if "&trainFile" ne "" %then %do;
            %let _trainFile = %superq(trainFile);
            %if ^%sysfunc(fileexist(%superq(_trainfile))) %then %do;
                data _null_;
                           putlog "WARNING: (&mname) %sysfunc(sasmsg(sashelp.aacomp, model_trainingfilenotfound_error, NOQUOTE, &_trainfile))";
                run;
            %end;
         %end;

        %let _spkFolder=;
        %if "&spkFolder" ne "" %then %let _spkFolder = %superq(spkFolder);
        %aa_createSPK(modelname=%superq(_modelName), modeldesc=%superq(_modeldesc), modelkey=&_modelKey, spkFolder=%superq(_spkFolder),
                inputVarTable= &_inputVarTable, outputVarTable=&_outputVarTable, tableMeta=&_tableMeta, targetVarTable=&_targetMeta,
                scorecodeFile=%superq(_scoreCodeFile), pmmlFile=%superq(_pmmlFile), trainFile=%superq(_trainFile), fitstats=&_fitstats,
                scorecodeFormat=&_scoreCodeFormat, miningFunction=&_miningFunction, miningAlgorithm=&_miningAlgorithm, itemstore=&_itemStoreFlag, debug=&_debug);
    %end;


   %if "%upcase(&debug)" eq "N" %then %do;
       proc datasets lib=work nowarn nolist;
          delete _model_fcmpDep;
          delete _model_inputVarsTable;
          delete _model_outputVarsTable;
          delete _model_targetMeta;
          delete _model_tableMeta;
          delete _model_varsmeta;
          delete _aaitemStoreVars;
          delete _aaitemStoreInfo;
          delete _aafitstats;
       run;
       quit;
  %end;

  %exit:

%mend aa_model_register;

%mend aaModel;
